# coding: utf-8

from hysop import __DEBUG__, __VERBOSE__, vprint, dprint
from hysop.tools.decorators import debug
from hysop.tools.types import to_list, to_set, to_tuple, first_not_None, check_instance
from hysop.tools.string_utils import framed_str, strlen, multiline_split
from hysop.tools.numpywrappers import npw
from hysop.core.graph.graph import not_implemented, initialized, discretized, \
    ready, graph_built, not_initialized
from hysop.core.graph.graph import ComputationalGraphNodeData
from hysop.core.graph.computational_node import ComputationalGraphNode
from hysop.core.graph.computational_operator import ComputationalGraphOperator
from hysop.core.graph.node_generator import ComputationalGraphNodeGenerator
from hysop.core.graph.node_requirements import NodeRequirements, OperatorRequirements
from hysop.core.memory.memory_request import MultipleOperatorMemoryRequests
from hysop.fields.field_requirements import MultiFieldRequirements
from hysop.topology.topology import Topology
from hysop.core.mpi import main_rank

from abc import ABCMeta, abstractmethod


class ComputationalGraph(ComputationalGraphNode):
    """
    Interface of an abstract graph of continuous operators (ie. a computational graph).
    """

    __metaclass__ = ABCMeta

    __FORCE_REPORTS__ = False

    @debug
    def __init__(self, candidate_input_tensors=None,
                 candidate_output_tensors=None,
                 **kwds):
        """
        Parameters
        ----------
        kwds: arguments for base classes.

        Notes
        -----
        The following base class variables cannot be specified during graph construction:
            variables, input_variables, output_variables
        Order of operation is: add_node, initialize, discretize,
                               get_work_properties, setup, apply, finalize.
        Nodes can also be added during pre initialization step.
        Graph building is done at the end of the initialization step, after all internal
        nodes have been initialized.
        """

        if ('input_fields' in kwds.keys()) or ('output_fields' in kwds.keys()):
            msg = 'input_fields or output_fields parameters should not be used in {}, they are \
                    deduced during graph construction (building step).'.format(cls)
            raise ValueError(msg)
        if ('input_params' in kwds.keys()) or ('output_params' in kwds.keys()):
            msg = 'input_params or output_params parameters should not be used in {}, they are \
                    deduced during graph construction (building step).'.format(cls)
            raise ValueError(msg)

        super(ComputationalGraph, self).__init__(input_fields=None, output_fields=None,
                                                 **kwds)

        self.nodes = []
        self.graph = None
        self.graph_built = False
        self.graph_is_rendering = False
        self.candidate_input_tensors = set(first_not_None(candidate_input_tensors, ()))
        self.candidate_output_tensors = set(first_not_None(candidate_output_tensors, ()))

    @graph_built
    def _get_profiling_info(self):
        """Collect profiling informations of profiled attributes."""
        for node in self.nodes:
            self._profiler += node._profiler

    def node_requirements_report(self, requirements):
        values = [(u'OPERATOR', u'TOPOLOGY', u'TSTATE', u'GHOSTS', u'MEMORY ORDER',
                   u'NODE.MRO[0]', u'NODE.MRO[1]', u'NODE.MRO[2]')]
        for node in self.nodes:
            reqs = node.get_node_requirements()
            if not isinstance(reqs, OperatorRequirements):
                continue
            opname = node.pretty_name.decode('utf-8')
            optypes = type(node).__mro__
            n = len(optypes)
            optypes = tuple(_.__name__ for _ in optypes[:min(3, n)]) + (u'',)*(3-n)
            vals = (opname,
                    reqs.enforce_unique_topology_shape, reqs.enforce_unique_transposition_state,
                    reqs.enforce_unique_ghosts, reqs.enforce_unique_memory_order,
                    ) + optypes
            vals = tuple(unicode(x) for x in vals)
            values.append(vals)

        template = u'\n   {:<{name_size}}   {:^{topology_size}}      {:^{tstates_size}}      {:^{ghosts_size}}      {:^{order_size}}      {:<{type_size0}}      {:<{type_size1}}      {:<{type_size2}}'
        name_size = max(strlen(s[0]) for s in values)
        topology_size = max(strlen(s[1]) for s in values)
        tstates_size = max(strlen(s[2]) for s in values)
        ghosts_size = max(strlen(s[3]) for s in values)
        order_size = max(strlen(s[4]) for s in values)
        type_size0 = max(strlen(s[5]) for s in values)
        type_size1 = max(strlen(s[6]) for s in values)
        type_size2 = max(strlen(s[7]) for s in values)

        ss = u''
        for (opname,  enforce_unique_topology_shape, enforce_unique_transposition_state,
                enforce_unique_ghosts, enforce_unique_memory_order, optype0, optype1, optype2) in values:
            ss += template.format(
                opname,
                enforce_unique_topology_shape,
                enforce_unique_transposition_state,
                enforce_unique_ghosts,
                enforce_unique_memory_order,
                optype0, optype1, optype2,
                name_size=name_size,
                topology_size=topology_size,
                tstates_size=tstates_size,
                ghosts_size=ghosts_size,
                order_size=order_size,
                type_size0=type_size0,
                type_size1=type_size1,
                type_size2=type_size2)

        title = u'ComputationalGraph {} node requirements report '.format(
            self.pretty_name.decode('utf-8'))
        return u'\n{}\n'.format(framed_str(title=title, msg=ss[1:])).encode('utf-8')

    def field_requirements_report(self, requirements):
        inputs, outputs = {}, {}
        sinputs, soutputs = {}, {}

        def sorted_reqs(reqs):
            return sorted(reqs, key=lambda x:
                          '{}::{}'.format(x.field.name, x.operator.name))
        for field, mreqs in requirements.input_field_requirements.iteritems():
            for td, reqs in mreqs.requirements.iteritems():
                for req in reqs:
                    inputs.setdefault(td, {}).setdefault(field, []).append(req)
        for td, td_reqs in inputs.iteritems():
            sin = sinputs.setdefault(td, [])
            for field, reqs in td_reqs.iteritems():
                for req in sorted_reqs(reqs):
                    opname = getattr(req.operator, 'pretty_name', 'UnknownOperator').decode('utf-8')
                    fname = getattr(req.field,    'pretty_name', 'UnknownField').decode('utf-8')
                    min_ghosts = req.ghost_str(req.min_ghosts)
                    max_ghosts = req.ghost_str(req.max_ghosts+1)
                    discr = str(req.operator.input_fields[field].grid_resolution)
                    ghosts = u'{}<=ghosts<{}'.format(min_ghosts, max_ghosts)
                    can_split = req.can_split.view(npw.int8)
                    memory_order = u'{}'.format(req.memory_order) if req.memory_order else u'ANY'
                    can_split = u'[{}]'.format(
                        u','.join('1' if cs else '0' for cs in req.can_split))
                    tstates = u'{}'.format(u','.join(str(ts) for ts in req.tstates)) \
                        if req.tstates else 'ANY'
                    sin.append((opname, fname, discr, ghosts, memory_order, can_split, tstates))
        for field, mreqs in requirements.output_field_requirements.iteritems():
            for td, reqs in mreqs.requirements.iteritems():
                for req in reqs:
                    outputs.setdefault(td, {}).setdefault(field, []).append(req)
        for td, td_reqs in outputs.iteritems():
            sout = soutputs.setdefault(td, [])
            for field, reqs in td_reqs.iteritems():
                for req in sorted_reqs(reqs):
                    opname = getattr(req.operator, 'pretty_name', 'UnknownOperator').decode('utf-8')
                    fname = getattr(req.field,    'pretty_name', 'UnknownField').decode('utf-8')
                    min_ghosts = req.ghost_str(req.min_ghosts)
                    max_ghosts = req.ghost_str(req.max_ghosts+1)
                    discr = str(req.operator.output_fields[field].grid_resolution)
                    ghosts = u'{}<=ghosts<{}'.format(min_ghosts, max_ghosts)
                    can_split = req.can_split.view(npw.int8)
                    memory_order = u'{}'.format(req.memory_order) if req.memory_order else u'ANY'
                    can_split = u'[{}]'.format(
                        u','.join('1' if cs else '0' for cs in req.can_split))
                    tstates = u'{}'.format(u','.join(str(ts) for ts in req.tstates)) \
                        if req.tstates else u'ANY'
                    sout.append((opname, fname, discr, ghosts, memory_order, can_split, tstates))

        titles = [[(u'OPERATOR', u'FIELD', u'DISCRETIZATION', u'GHOSTS',
                    u'MEMORY ORDER', u'CAN_SPLIT', u'TSTATES')]]
        name_size = max(len(s[0]) for ss in sinputs.values()+soutputs.values()+titles for s in ss)
        field_size = max(len(s[1]) for ss in sinputs.values()+soutputs.values()+titles for s in ss)
        discr_size = max(len(s[2]) for ss in sinputs.values()+soutputs.values()+titles for s in ss)
        ghosts_size = max(len(s[3]) for ss in sinputs.values()+soutputs.values()+titles for s in ss)
        order_size = max(len(s[4]) for ss in sinputs.values()+soutputs.values()+titles for s in ss)
        cansplit_size = max(len(s[5]) for ss in sinputs.values() +
                            soutputs.values()+titles for s in ss)
        tstates_size = max(len(s[6]) for ss in sinputs.values() +
                           soutputs.values()+titles for s in ss)

        template = u'\n   {:<{name_size}}   {:^{field_size}}     {:^{discr_size}}      {:^{ghosts_size}}      {:^{order_size}}      {:^{cansplit_size}}      {:^{tstates_size}}'

        ss = u'>INPUTS:'
        if sinputs:
            for (td, sreqs) in sinputs.iteritems():
                if isinstance(td, Topology):
                    ss += u'\n {}'.format(td.short_description())
                else:
                    ss += u'\n {}'.format(td)
                ss += template.format(*titles[0][0],
                                      name_size=name_size, field_size=field_size,
                                      discr_size=discr_size, ghosts_size=ghosts_size,
                                      order_size=order_size, cansplit_size=cansplit_size,
                                      tstates_size=tstates_size)
                for (opname, fname, discr, ghosts, order, can_split, tstates) in sreqs:
                    ss += template.format(
                        opname, fname, discr, ghosts, order, can_split, tstates,
                        name_size=name_size, field_size=field_size,
                        discr_size=discr_size, ghosts_size=ghosts_size,
                        order_size=order_size, cansplit_size=cansplit_size,
                        tstates_size=tstates_size)
        else:
            ss += u' None'
        ss += u'\n>OUTPUTS:'
        if soutputs:
            for (td, sreqs) in soutputs.iteritems():
                if isinstance(td, Topology):
                    ss += u'\n {}'.format(td.short_description())
                else:
                    ss += u'\n {}'.format(td)
                ss += template.format(*titles[0][0],
                                      name_size=name_size, field_size=field_size,
                                      discr_size=discr_size, ghosts_size=ghosts_size,
                                      order_size=order_size, cansplit_size=cansplit_size,
                                      tstates_size=tstates_size)
                for (opname, fname, discr, ghosts, order, can_split, tstates) in sreqs:
                    ss += template.format(
                        opname, fname, discr, ghosts, order, can_split, tstates,
                        name_size=name_size, field_size=field_size,
                        discr_size=discr_size, ghosts_size=ghosts_size,
                        order_size=order_size, cansplit_size=cansplit_size,
                        tstates_size=tstates_size)
        else:
            ss += u' None'

        title = u'ComputationalGraph {} field requirements report '.format(
            self.pretty_name.decode('utf-8'))
        return u'\n{}\n'.format(framed_str(title=title, msg=ss)).encode('utf-8')

    def domain_report(self):
        domains = self.get_domains()
        ops = {}

        maxlen = (None, 40,  None, 40,  None)
        split_sep = (None, ',', None, ',', None)
        newline_prefix = (None, ' ', '',   ' ', None)
        replace = ('',   '', '-',   '',  '')

        for (domain, operators) in domains.iteritems():
            if (domain is None):
                continue
            for op in sorted(operators, key=lambda x: x.pretty_name):
                finputs = u','.join(sorted([f.pretty_name.decode('utf-8')
                                            for f in op.iter_input_fields() if f.domain is domain]))
                foutputs = u','.join(sorted([f.pretty_name.decode('utf-8')
                                             for f in op.iter_output_fields() if f.domain is domain]))
                pinputs = u','.join(sorted([p.pretty_name.decode('utf-8')
                                            for p in op.input_params.values()]))
                poutputs = u','.join(sorted([p.pretty_name.decode('utf-8')
                                             for p in op.output_params.values()]))
                infields = u'[{}]'.format(finputs) if finputs else u''
                outfields = u'[{}]'.format(foutputs) if foutputs else u''
                inparams = u'[{}]'.format(pinputs) if pinputs else u''
                outparams = u'[{}]'.format(poutputs) if poutputs else u''

                inputs = u'{}{}{}'.format(
                    infields,  u'x' if infields and inparams else u'', inparams)
                outputs = u'{}{}{}'.format(
                    outfields, u'x' if outfields and outparams else u'', outparams)

                if inputs == u'':
                    inputs = u'no inputs'
                if outputs == u'':
                    outputs = u'no outputs'

                opname = op.pretty_name.decode('utf-8')
                optype = type(op).__name__
                strdata = (opname, inputs, '->', outputs, optype)

                op_data = ops.setdefault(domain, [])
                op_data += multiline_split(strdata, maxlen, split_sep, replace, newline_prefix)

        if (None in domains):
            operators = domains[None]
            for op in sorted(operators, key=lambda x: x.pretty_name):
                pinputs = u','.join(sorted([p.pretty_name.decode('utf-8')
                                            for p in op.input_params.values()]))
                poutputs = u','.join(sorted([p.pretty_name.decode('utf-8')
                                             for p in op.output_params.values()]))
                inparams = u'[{}]'.format(pinputs) if pinputs else ''
                outparams = u'[{}]'.format(poutputs) if poutputs else ''

                inputs = u'{}'.format(inparams)
                outputs = u'{}'.format(outparams)
                if inputs == '':
                    inputs = u'no inputs'
                if outputs == '':
                    outputs = u'no outputs'
                opname = op.pretty_name.decode('utf-8')
                optype = type(op).__name__
                strdata = (opname, inputs, '->', outputs, optype)

                op_data = ops.setdefault(None, [])
                op_data += multiline_split(strdata, maxlen, split_sep, replace, newline_prefix)

        name_size = max(strlen(s[0]) for ss in ops.values() for s in ss)
        in_size = max(strlen(s[1]) for ss in ops.values() for s in ss)
        arrow_size = max(strlen(s[2]) for ss in ops.values() for s in ss)
        out_size = max(strlen(s[3]) for ss in ops.values() for s in ss)
        type_size = max(strlen(s[4]) for ss in ops.values() for s in ss)

        ss = u''
        for (domain, dops) in ops.iteritems():
            if (domain is None):
                continue
            ss += u'\n>{}'.format(domain.short_description())
            ss += u'\n   {:<{name_size}}  {:<{in_size}}  {:<{arrow_size}}   {:<{out_size}}    {:<{type_size}}'.format(
                'OPERATOR', 'INPUTS', '', 'OUTPUTS', 'OPERATOR TYPE',
                name_size=name_size, in_size=in_size,
                arrow_size=arrow_size,
                out_size=out_size, type_size=type_size)
            for (opname, inputs, arrow, outputs, optype) in dops:
                ss += u'\n   {:<{name_size}}  {:<{in_size}}  {:<{arrow_size}}   {:<{out_size}}    {:<{type_size}}'.format(
                    opname, inputs, arrow, outputs, optype,
                    name_size=name_size, in_size=in_size,
                    arrow_size=arrow_size,
                    out_size=out_size, type_size=type_size)
        if (None in domains):
            ss += u'\n>Domainless operators:'
            ss += u'\n   {:<{name_size}}  {:<{in_size}}  {:<{arrow_size}}   {:<{out_size}}    {:<{type_size}}'.format(
                'OPERATOR', 'INPUTS', '', 'OUTPUTS', 'OPERATOR TYPE',
                name_size=name_size, in_size=in_size,
                arrow_size=arrow_size,
                out_size=out_size, type_size=type_size)
            for (opname, inputs, arrow, outputs, optype) in ops[None]:
                ss += u'\n   {:<{name_size}}  {:<{in_size}}  {:<{arrow_size}}   {:<{out_size}}    {:<{type_size}}'.format(
                    opname, inputs, arrow, outputs, optype,
                    name_size=name_size, in_size=in_size,
                    arrow_size=arrow_size,
                    out_size=out_size, type_size=type_size)

        title = u'ComputationalGraph {} domain and operator report '.format(
            self.pretty_name.decode('utf-8'))
        return u'\n{}\n'.format(framed_str(title=title, msg=ss[1:])).encode('utf-8')

    def topology_report(self):
        ss = ''
        for (backend, topologies) in self.get_topologies().iteritems():
            ss += u'\n {}:'.format(backend.short_description())
            ss += u'\n  *'+'\n  *'.join(t.short_description()
                                        for t in sorted(topologies, key=lambda x: x.id))
        title = u'ComputationalGraph {} topology report '.format(
            self.pretty_name.decode('utf-8'))
        return u'\n{}\n'.format(framed_str(title=title, msg=ss[1:]))

    def variable_report(self):
        fields = self.fields

        topologies = {}
        for field in self.fields:
            field_topologies = {}
            for (i, node) in enumerate(self.nodes):
                if field in node.input_fields:
                    topo = node.input_fields[field]
                    field_topologies.setdefault(topo, []).append(node)
                if field in node.output_fields:
                    topo = node.output_fields[field]
                    field_topologies.setdefault(topo, []).append(node)
            for topo in sorted(field_topologies.keys(), key=lambda x: x.tag):
                pnames = set(node.pretty_name.decode('utf-8') for node in field_topologies[topo])
                pnames = sorted(pnames)
                nbyline = 4
                nentries = len(pnames)//nbyline
                n0 = len(str(topo.backend.kind).lower())
                n1 = len(str(topo.tag))
                for i in xrange(nentries):
                    sops = u', '.join(pnames[nbyline*i:nbyline*(i+1)])
                    if (i != nentries-1) or (len(pnames) % nbyline != 0):
                        sops += ','
                    if (i == 0):
                        entries = (str(topo.backend.kind).lower(), topo.tag, sops)
                    else:
                        entries = ('', '-'*n1, sops)
                    topologies.setdefault(field, []).append(entries)
                if (len(pnames) % nbyline != 0):
                    sops = u', '.join(pnames[nbyline*nentries:])
                    if (nentries == 0):
                        entries = (str(topo.backend.kind).lower(), topo.tag, sops)
                    else:
                        entries = ('', '-'*n1, sops)
                    topologies.setdefault(field, []).append(entries)

        titles = [[(u'BACKEND', u'TOPOLOGY', u'OPERATORS')]]
        backend_size = max(len(s[0]) for ss in topologies.values()+titles for s in ss)
        topo_size = max(len(s[1]) for ss in topologies.values()+titles for s in ss)
        template = u'\n   {:<{backend_size}}   {:<{topo_size}}   {}'
        sizes = {'backend_size': backend_size,
                 'topo_size': topo_size}

        ss = u''
        for field in sorted(self.fields, key=lambda x: x.name):
            ss += u'\n>FIELD {}::{}'.format(field.name, field.pretty_name.decode('utf-8'))
            ss += template.format(*titles[0][0], **sizes)
            field_topologies = topologies[field]
            for entries in field_topologies:
                ss += template.format(*entries, **sizes)

        title = u'ComputationalGraph {} fields report '.format(
            self.pretty_name.decode('utf-8'))
        ss = u'\n{}\n'.format(framed_str(title=title, msg=ss[1:]))
        return ss.encode('utf-8')

    def operator_report(self):
        maxlen = (None, None, 40,  None, 40,  None)
        split_sep = (None, None, ',', None, ',', None)
        newline_prefix = (None, None, ' ', '',   ' ', None)
        replace = ('--',  '',  '',  '-',  '',  '')

        reduced_graph = self.reduced_graph
        ops = []
        for (i, node) in enumerate(self.nodes):
            handled_inputs, handled_outputs = (), ()
            finputs, foutputs = [], []
            for f in node.input_tensor_fields:
                f0 = f.fields[0]
                t0 = node.input_fields[f0]
                if all((node.input_fields[fi] is t0) for fi in f.fields):
                    finputs.append(u'{}.{}'.format(f.pretty_name.decode('utf-8'),
                                                   t0.pretty_tag.decode('utf-8')))
                    handled_inputs += f.fields
            for f in node.output_tensor_fields:
                f0 = f.fields[0]
                t0 = node.output_fields[f0]
                if all((node.output_fields[fi] is t0) for fi in f.fields):
                    foutputs.append(u'{}.{}'.format(f.pretty_name.decode('utf-8'),
                                                    t0.pretty_tag.decode('utf-8')))
                    handled_outputs += f.fields
            finputs += [u'{}.{}'.format(f.pretty_name.decode('utf-8'),
                                        t.pretty_tag.decode('utf-8'))
                        for (f, t) in node.input_fields.iteritems()
                        if f not in handled_inputs]
            foutputs += [u'{}.{}'.format(f.pretty_name.decode('utf-8'),
                                         t.pretty_tag.decode('utf-8'))
                         for (f, t) in node.output_fields.iteritems()
                         if f not in handled_outputs]
            finputs = u','.join(sorted(finputs))
            foutputs = u','.join(sorted(foutputs))

            pinputs = u','.join(sorted([p.pretty_name.decode('utf-8')
                                        for p in node.input_params.values()]))
            poutputs = u','.join(sorted([p.pretty_name.decode('utf-8')
                                         for p in node.output_params.values()]))

            infields = u'[{}]'.format(finputs) if finputs else u''
            outfields = u'[{}]'.format(foutputs) if foutputs else u''
            inparams = u'[{}]'.format(pinputs) if pinputs else u''
            outparams = u'[{}]'.format(poutputs) if poutputs else u''

            inputs = u'{}{}{}'.format(infields,  u'x' if infields and inparams else u'', inparams)
            outputs = u'{}{}{}'.format(
                outfields, u'x' if outfields and outparams else u'', outparams)
            if inputs == '':
                inputs = u'no inputs'
            if outputs == '':
                outputs = u'no outputs'

            opname = node.pretty_name.decode('utf-8')
            optype = type(node).__name__
            strdata = (str(i), opname, inputs, '->', outputs, optype)

            ops += multiline_split(strdata, maxlen, split_sep, replace, newline_prefix)

        isize = max(strlen(s[0]) for s in ops)
        name_size = max(strlen(s[1]) for s in ops)
        in_size = max(strlen(s[2]) for s in ops)
        arrow_size = max(strlen(s[3]) for s in ops)
        out_size = max(strlen(s[4]) for s in ops)
        type_size = max(strlen(s[5]) for s in ops)

        ss = u'  {:<{isize}}  {:<{name_size}}  {:<{in_size}}  {:<{arrow_size}}   {:<{out_size}}    {:<{type_size}}'.format(
            'ID', 'OPERATOR', 'INPUTS', '', 'OUTPUTS', 'OPERATOR TYPE',
            isize=isize,
            name_size=name_size, in_size=in_size,
            arrow_size=arrow_size,
            out_size=out_size, type_size=type_size)
        for (i, opname, inputs, arrow, outputs, optype) in ops:
            ss += u'\n  {:>{isize}}  {:<{name_size}}  {:<{in_size}}  {:<{arrow_size}}   {:<{out_size}}    {:<{type_size}}'.format(
                i, opname, inputs, arrow, outputs, optype,
                isize=isize,
                name_size=name_size, in_size=in_size,
                arrow_size=arrow_size,
                out_size=out_size, type_size=type_size)

        title = u'ComputationalGraph {} discrete operator report '.format(
            self.pretty_name.decode('utf-8'))
        return u'\n{}\n'.format(framed_str(title=title, msg=ss)).encode('utf-8')

    def get_domains(self):
        domains = {}
        for node in self.nodes:
            for (domain, ops) in node.get_domains().iteritems():
                domains.setdefault(domain, set()).update(ops)
        return domains

    @graph_built
    def get_topologies(self):
        topologies = {}
        for node in self.nodes:
            node_topologies = node.get_topologies()
            for (backend, topos) in node_topologies.iteritems():
                topos = to_set(topos)
                if backend not in topologies:
                    topologies[backend] = set()
                topologies[backend].update(topos)
        return topologies

    @debug
    @not_initialized
    def push_nodes(self, *args):
        nodes = ()
        for arg in args:
            if (arg is None):
                continue
            nodes += to_tuple(arg)
        for node in nodes:
            if isinstance(node, ComputationalGraph):
                self.candidate_input_tensors.update(node.candidate_input_tensors)
                self.candidate_output_tensors.update(node.candidate_output_tensors)
                self.nodes.append(node)
            elif isinstance(node, ComputationalGraphNode):
                self.candidate_input_tensors.update(node.input_tensor_fields)
                self.candidate_output_tensors.update(node.output_tensor_fields)
                self.nodes.append(node)
            elif isinstance(node, ComputationalGraphNodeGenerator):
                self.candidate_input_tensors.update(node.candidate_input_tensors)
                self.candidate_output_tensors.update(node.candidate_output_tensors)
                nodes = node.generate()
                assert (nodes is not None), node
                self.push_nodes(*nodes)
            else:
                msg = 'Given node is not an instance of ComputationalGraphNode (got a {}).'
                raise ValueError(msg.format(node.__class__))
        return self

    def available_methods(self):
        avail_methods = {}
        if not self.nodes:
            msg = u'No nodes present in ComputationalGraph {}.'.format(
                self.pretty_name.decode('utf-8'))
            raise RuntimeError(msg.encode('utf-8'))
        for node in self.nodes:
            for (k, v) in node.available_methods().iteritems():
                v = to_set(v)
                if (k in avail_methods):
                    avail_methods[k].update(v)
                else:
                    avail_methods[k] = v
        return avail_methods

    def default_method(self):
        return {}

    @debug
    def initialize(self,
                   is_root=True,
                   topgraph_method=None,
                   outputs_are_inputs=False,
                   **kwds):
        if self.initialized:
            return
        self.is_root = is_root

        if is_root:
            self.pre_initialize(**kwds)

        msg = u'ComputationalGraph {} is empty.'
        assert len(self.nodes) > 0, msg.format(self.pretty_name.decode('utf-8')).encode('utf-8')

        for node in self.nodes:
            node.pre_initialize(**kwds)

        method = self._setup_method(topgraph_method)
        self.handle_method(method)

        for node in self.nodes:
            node.initialize(is_root=False,
                            outputs_are_inputs=False,
                            topgraph_method=method,
                            **kwds)

        for node in self.nodes:
            node.post_initialize(**kwds)

        if (is_root and __VERBOSE__) or __DEBUG__ or self.__FORCE_REPORTS__:
            print self.domain_report()

        if is_root:
            field_requirements = self.get_and_set_field_requirements()
            field_requirements.build_topologies()

            self._build_graph(outputs_are_inputs=outputs_are_inputs, current_level=0)

            # fix for auto generated nodes
            for node in self.nodes:
                if (node._field_requirements is None):
                    node.get_and_set_field_requirements()
                    assert (node._field_requirements is not None)

            # from now on, all nodes contained in self.nodes are ComputationalGraphOperator
            # (ordered for sequential execution)
            self.handle_topologies(self.input_topology_states, self.output_topology_states)
            self.check()

        self.initialized = True
        if is_root:
            self.post_initialize(**kwds)

    @debug
    def check(self):
        super(ComputationalGraph, self).check()
        reduced_graph = self.reduced_graph
        for node in self.nodes:
            node.check()

    @debug
    def get_field_requirements(self):
        requirements = super(ComputationalGraph, self).get_field_requirements()
        if (self.is_root and __VERBOSE__) or __DEBUG__ or self.__FORCE_REPORTS__:
            print self.node_requirements_report(requirements)
        for node in self.nodes:
            node_requirements = node.get_and_set_field_requirements()
            requirements.update(node_requirements)
        if (self.is_root and __VERBOSE__) or __DEBUG__ or self.__FORCE_REPORTS__:
            print self.field_requirements_report(requirements)
        return requirements

    @debug
    def handle_topologies(self, input_topology_states, output_topology_states):
        from hysop.problem import Problem
        # do not call super method
        for node in self.nodes:
            assert isinstance(node, ComputationalGraphOperator) or isinstance(node, Problem)
            assert (node in input_topology_states)
            assert (node in output_topology_states)
            assert (node._field_requirements is not None)
            if not isinstance(node, Problem):
                node.handle_topologies(input_topology_states[node], output_topology_states[node])
            node.input_topology_states = input_topology_states[node]
            node.output_topology_states = output_topology_states[node]
        self.topology_handled = True

    @debug
    def _build_graph(self, current_level, outputs_are_inputs=False, **kwds):
        if self.graph_built:
            return

        from hysop.core.graph.graph_builder import GraphBuilder

        builder = GraphBuilder(node=self)
        builder.configure(current_level=current_level,
                          outputs_are_inputs=outputs_are_inputs, **kwds)
        builder.build_graph()

        input_fields = builder.input_fields
        output_fields = builder.output_fields
        candidate_input_tensors = self.candidate_input_tensors
        candidate_output_tensors = self.candidate_output_tensors

        input_tensor_fields = ()
        output_tensor_fields = ()
        for itfield in candidate_input_tensors:
            if all((f in input_fields) for f in itfield):
                input_tensor_fields += (itfield,)
        for otfield in candidate_output_tensors:
            if all((f in output_fields) for f in otfield):
                output_tensor_fields += (otfield,)

        # keep variables
        self._init_base(input_fields=input_fields,
                        output_fields=output_fields,
                        input_tensor_fields=input_tensor_fields,
                        output_tensor_fields=output_tensor_fields,
                        input_params=builder.input_params,
                        output_params=builder.output_params)

        self.graph = builder.graph
        self.reduced_graph = builder.reduced_graph
        self.sorted_nodes = builder.sorted_nodes
        self.nodes = builder.nodes
        self.input_topology_states = builder.op_input_topology_states
        self.output_topology_states = builder.op_output_topology_states

        self.initial_input_topology_states = builder.input_topology_states
        self.final_output_topology_states = builder.output_topology_states

        self.level = current_level

        self.initialized = True
        self.graph_built = True

        if (self.is_root and __VERBOSE__) or __DEBUG__ or self.__FORCE_REPORTS__:
            print self.topology_report()
            print self.variable_report()
            print self.operator_report()

    def display(self, visu_rank=0, show_buttons=False):
        """
        Display the reduced computational graph.
        """
        from hysop import main_rank
        if (visu_rank is None) or (main_rank != visu_rank):
            return
        
        net = self.to_pyvis() 
        
        import tempfile
        with tempfile.NamedTemporaryFile(suffix='.html') as f:
            net.show(f.name)
    
    def to_html(self, path, io_rank=0, show_buttons=False):
        """
        Generate an interactive computational graph in an html file.
        """
        from hysop import main_rank
        if (io_rank is None) or (main_rank != io_rank):
            return
        
        net = self.to_pyvis() 
        net.write_html(path)

    @graph_built
    def to_pyvis(self):
        """
        Convert the graph to a pyvis network for vizualization.
        """
        try:
            import pyvis
        except ImportError:
            msg='\nFATAL ERROR: Graph vizualization requires the pyvis module.\n'
            print(msg)
            raise

        graph = self.reduced_graph
        network = pyvis.network.Network(directed=True)
        known_nodes = set()

        def add_node(node):
            node_id = int(node)
            if node_id not in known_nodes:
                network.add_node(node_id, label=node.label, 
                        title=node.title, color=node.color)
                known_nodes.add(node_id)
        
        def add_edge(from_node, to_node):
            from_node_id = int(from_node)
            to_node_id   = int(to_node)
            edge = graph[from_node][to_node]
            title = edge.get('data', 'FAIL')
            network.add_edge(from_node_id, to_node_id, title=title)

        for node in graph:
            add_node(node)
            for out_node in graph[node]:
                add_node(out_node)
                add_edge(node, out_node)

        return network


    @debug
    @graph_built
    def discretize(self):
        if self.discretized:
            return
        for node in self.nodes:
            if not node.discretized:
                node.discretize()

        if self.is_root:
            input_discrete_fields = {}
            for (field, topo) in self.input_fields.iteritems():
                istate = self.initial_input_topology_states[field][1]
                # problem inputs are writeable for initialization
                istate = istate.copy(is_read_only=False)
                dfield = field.discretize(topo, istate)
                input_discrete_fields[field] = dfield

            output_discrete_fields = {}
            for field, topo in self.output_fields.iteritems():
                ostate = self.final_output_topology_states[field][1]
                dfield = field.discretize(topo, ostate)
                output_discrete_fields[field] = dfield

            # build back DiscreteTensorFields from DiscreteScalarFields
            from hysop.fields.discrete_field import DiscreteTensorField
            input_discrete_tensor_fields = {}
            for tfield in self.input_tensor_fields:
                dfields = tfield.new_empty_array()
                for (idx, field) in tfield.nd_iter():
                    dfield = input_discrete_fields[field]
                    dfields[idx] = dfield
                tdfield = DiscreteTensorField(field=tfield, dfields=dfields)
                input_discrete_tensor_fields[tfield] = tdfield

            output_discrete_tensor_fields = {}
            for tfield in self.output_tensor_fields:
                dfields = tfield.new_empty_array()
                for (idx, field) in tfield.nd_iter():
                    dfield = output_discrete_fields[field]
                    dfields[idx] = dfield
                tdfield = DiscreteTensorField(field=tfield, dfields=dfields)
                output_discrete_tensor_fields[tfield] = tdfield

            discrete_fields = tuple(set(input_discrete_fields.values() +
                                        output_discrete_fields.values()))

            discrete_tensor_fields = tuple(set(input_discrete_tensor_fields.values() +
                                               output_discrete_tensor_fields.values()))

        else:
            input_discrete_fields = None
            output_discrete_fields = None
            input_discrete_tensor_fields = None
            output_discrete_tensor_fields = None
            discrete_fields = None
            discrete_tensor_fields = None

        self.input_discrete_fields = input_discrete_fields
        self.output_discrete_fields = output_discrete_fields
        self.input_discrete_tensor_fields = input_discrete_tensor_fields
        self.output_discrete_tensor_fields = output_discrete_tensor_fields
        self.discrete_fields = discrete_fields
        self.discrete_tensor_fields = discrete_tensor_fields

        self.discretized = True

    @debug
    @discretized
    def get_work_properties(self):
        requests = MultipleOperatorMemoryRequests()

        for node in self.nodes:
            if node not in requests.operators():
                wp = node.get_work_properties()
                requests += node.get_work_properties()
        if __DEBUG__ or (__VERBOSE__ and self.level == 0) or self.__FORCE_REPORTS__:
            srequests = requests.sreport()
            ss = (srequests if (srequests != u'') else u' *no extra work requested*')
            title = u'ComputationalGraph {} work properties report '.format(
                self.pretty_name.decode('utf-8'))
            vprint(u'\n{}\n'.format(framed_str(title=title, msg=ss)).encode('utf-8'))
        return requests

    @debug
    @discretized
    def setup(self, work=None, allow_subbuffers=False):
        if self.ready:
            return
        if (work is None):
            work = self.get_work_properties()
            work.allocate(allow_subbuffers=allow_subbuffers)
        for node in self.nodes:
            if not node.ready:
                node.setup(work=work)
        self.ready = True

    def build(self, outputs_are_inputs=True, method=None, allow_subbuffers=False):
        """
        Shortcut for initialize(), discretize(), get_work_properties(), setup()
        for quick graph initialization.
        """
        self.initialize(outputs_are_inputs=outputs_are_inputs,
                        topgraph_method=method)
        assert self.is_root, 'Only root graph can be built.'
        self.discretize()
        work = self.get_work_properties()
        work.allocate(allow_subbuffers=allow_subbuffers)
        self.setup(work)
        return self

    @debug
    @ready
    def apply(self, **kwds):
        for node in self.nodes:
            dprint('{}.apply()'.format(node.name))
            node.apply(**kwds)

    @debug
    @ready
    def finalize(self, **kwds):
        reduced_graph = self.reduced_graph
        for node in self.nodes:
            if node.ready:
                node.finalize(**kwds)
        self.ready = False

    @classmethod
    def supports_multiple_field_topologies(cls):
        return True

    @classmethod
    def supports_multiple_topologies(cls):
        return True

    @classmethod
    def supports_mpi(cls):
        return True
