"""Abstract class providing a common interface to all
discrete operators working on GPU.

* :class:`~hysop.gpu_operator.GPUOperator` is an abstract class
    used to provide a common interface to all discrete operators
    working on GPU.
    See for example :class:`~hysop.gpu.gpu_diffusion.GPUDiffusion` or
    :class:`~hysop.gpu.gpu_particle_advection.GPUParticleAdvection`.

"""
from abc import ABCMeta
from hysop.constants import HYSOP_REAL, DirectionLabels
from hysop.backend.device.opencl.opencl_tools import get_opencl_environment
from hysop.methods import ExtraArgs
from hysop.operator.discrete.discrete import get_extra_args_from_method
from hysop.backend.device.kernel_config import KernelConfig
from hysop.methods import Precision

class GPUOperator(object):
    """Abstract class for discrete operators working on GPU.
    """
    __metaclass__ = ABCMeta

    def __init__(self, platform_id=None, device_id=None, device_type=None,
                 direction=None, gpu_precision=HYSOP_REAL, **kwds):
        """
        Create the common attributes of all GPU discrete operators.
        All in this interface is independant of a discrete operator.

        Parameters
        ----------
        direction : int
            leading direction of work (for example advection dir)
        platform_id : int, optional
            OpenCL id, default = 0.
        device_id : int, optional
            OpenCL id, default = 0.
        device_type : string, optional.
            OpenCL selected device, default = 'gpu'.
        gpu_precision : numpy.dtype, optional
            floating precision for gpu, default=HYSOP_REAL
        """
        
        super(GPUOperator, self).__init__(**kwds)

        ## real type precision on GPU
        self.gpu_precision = gpu_precision
        if 'method' in kwds:
            method = kwds['method']
            assert (method is not None)
            
            if Precision in method:
                self.gpu_precision = method[Precision]
            
            if KernelConfig in method:
                config = method[KernelConfig]
                platform_id = config.platform_id
                device_id   = config.device_id
                device_type = config.device_type
            elif ExtraArgs in method:
                platform_id = get_extra_args_from_method(self,'platform_id',platform_id)
                device_id   = get_extra_args_from_method(self,'device_id',device_id)
                device_type = get_extra_args_from_method(self,'device_type',device_type)

        # Initialize opencl environment
        comm_ref = self.variables[0].topology.comm

        self.cl_env = get_opencl_environment(
            platform_id=platform_id, device_id=device_id,
            device_type=device_type, precision=self.gpu_precision,
            comm=comm_ref)
        
        # In case of directional splitting, define a function
        # to get the appropriate vectors for the current provided direction.
        # Only define _reorder_vect if direction is provided to avoid bugs.
        dim = self.domain.dim
        if (direction is not None):
            if (direction<0) or (direction >= dim):
                msg = "Bad direction '{}' for dimension {}.".format(direction,dim)
                raise ValueError(msg)

            if   dim == 1:
                _reorder_vect = lambda v: (v[0],)
            elif dim == 2:
                if direction == 0:
                    _reorder_vect = lambda v: (v[0], v[1])
                elif direction == 1:
                    _reorder_vect = lambda v: (v[1], v[0])
            elif dim == 3: 
                if direction == 0:
                    _reorder_vect = lambda v: (v[0], v[1], v[2])
                elif direction == 1:
                    _reorder_vect = lambda v: (v[1], v[0], v[2])
                elif direction == 2:
                    _reorder_vect = lambda v: (v[2], v[0], v[1])
            else:
                msg = 'Dimension {} not implemented yet!'.format(dim)
                raise NotImplementedError(msg)
            
            self._reorder_vect = _reorder_vect
    
        self.dim = dim
        self.direction = direction

        # Size constants for local mesh size
        self._size_constants = ''
    
        # Try to load kernel configurations
        try:
            self._kernel_cfg = \
                self.cl_env.kernels_config[self.dim][self.gpu_precision]
        except:
            self._kernel_cfg = None

        # Global memory allocated on gpu by this operator
        self.size_global_alloc = 0
        # Local memory allocated on gpu by this operator
        self.size_local_alloc = 0
        
        self._num_locMem = None

    def _append_size_constants(self, values, prefix='NB', suffix=None):
        """Append to the string containing the constants for building kernels.

        Parameters
        ----------
        values : list
        prefix : string, optional
        suffix : list of strings, optional
           directions, default = `hysop.constants.DirectionLabels`.
        """
        if suffix is None:
            suffix = DirectionLabels
        assert len(values) <= len(suffix), str(values) + str(suffix)
        for v, s in zip(values, suffix):
            self._size_constants += " -D " + prefix + s + "=" + str(v)

