# -*- coding: utf-8 -*-
"""Library of functions used to perform classical vector calculus
(diff operations like grad, curl ...) based on finite differences schemes.

* :class:`~hysop.numerics.differential_operations.Curl`,
* :class:`~hysop.numerics.differential_operations.DivRhoV`,
* :class:`~hysop.numerics.differential_operations.DivWV`,
* :class:`~hysop.numerics.differential_operations.Laplacian`,
* :class:`~hysop.numerics.differential_operations.GradS`,
* :class:`~hysop.numerics.differential_operations.GradV`,
* :class:`~hysop.numerics.differential_operations.GradVxW`,
* :class:`~hysop.numerics.differential_operations.DivAdvection`,
* :class:`~hysop.numerics.differential_operations.Strain`,
* :class:`~hysop.numerics.differential_operations.StrainCriteria`,
* :class:`~hysop.numerics.differential_operations.MaxDiagGradV`,
* :class:`~hysop.numerics.differential_operations.StretchLike`,
* :class:`~hysop.numerics.differential_operations.DiagAndStretch`,
* :class:`~hysop.numerics.differential_operations.StrainCriteria`,
* :class:`~hysop.numerics.differential_operations.StrainAndStretch`,
* :class:`~hysop.numerics.differential_operations.DifferentialOperation`,
 (abstract base class).

Notes
-----

For coherence sake, all input/output variables used in differential operations
are list of numpy arrays, even when only scalar fields are required.

For example :

* a = curl(b) in 3d, a and b vector fields and so lists of 3 arrays.
* a = DivRhoV(c,d), d a vector field, a and c scalar fields. d is a list of
3 arrays and a and c lists of 1 array.

For practical examples of use, see test_differential_operations.py file.

"""
from hysop.constants import debug, XDIR, YDIR, ZDIR
from abc import ABCMeta
from hysop.numerics.finite_differences import FDC4, FDC2, FD2C2
import numpy as np
from hysop.tools.numpywrappers import npw
from hysop.tools.misc import WorkSpaceTools, Utils


class DifferentialOperation(object):
    """Abstract base class for all operations
    based on finite differences.
    """
    __metaclass__ = ABCMeta
    _authorized_methods = []

    # @debug
    # def __new__(cls, *args, **kw):
    #     return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, topo, indices=None, reduce_output_shape=False,
                 method=None, work=None):
        """
        Parameters
        ----------
        topo : :class:`hysop.topology.topology.CartesianTopology`
        indices : list of slices, optional
            Represents the local mesh on which the operation
            will be applied,
            like compute_index in :class:`~hysop.domain.mesh.Mesh`.
            See details in notes.
        reduce_output_shape : boolean, optional
            True to return the result in a reduced array. See notes below.
        method : :class:`hysop.numerics.finite_differences.FiniteDifference`
        , optional
            the chosen FD scheme. If None, a default method is set, equal
            to self._authorized_methods[0].
        work : list of numpy arrays, optional
            internal buffers

        Notes
        -----
        * Two ways to compute outvar = operation(invar, ...)
        ** 1 - either invar and outvar are arrays of the same shape, and
        outvar[...] = op(invar[...]) will be computed. In that case
        indices and reduce_output_shape are not required.
        Indices will be set to topo.mesh.compute_index.
        ** 2 - or outvar is a smaller array than invar, and
        outvar[...] = op(invar(indices))
        To use case 2, set reduce_output_shape = True and provide indices.
        * See ClassName._authorized_methods[0] to find which default method
        is set for each Operation.
        * If work = None, some work arrays will be allocated internally.
        Else, you must provide a list of lwk arrays of shape
        topo.mesh.local_resolution,

        """
        self._dim = topo.domain.dim
        # Set default method
        if method is None:
            method = self._authorized_methods[0]

        self.method = method
        if indices is None:
            indices = topo.mesh.compute_index
        self.in_indices = indices
        # True if work has to be done for this set
        # of indices on the current proc.
        self._on_proc = Utils.is_on_proc(indices)
        self.fd_scheme = self._init_fd_method(topo, reduce_output_shape)
        self.output_indices = self.fd_scheme.output_indices
        self._work = self._set_work_arrays(topo, reduce_output_shape, work)

    def _set_work_arrays(self, topo, reduce_output_shape, rwork=None):
        """Check and/or allocate internal work buffers.
        """
        wk_prop = self.get_work_properties(topo, self.in_indices)['rwork']
        if wk_prop is None:
            return []
        if reduce_output_shape:
            subshape = np.asarray([self.in_indices[i].stop -
                                   self.in_indices[i].start
                                   for i in xrange(self._dim)])
        else:
            subshape = topo.mesh.local_resolution
        subshape = tuple(subshape)
        return WorkSpaceTools.check_work_array(len(wk_prop), subshape, rwork)

    @staticmethod
    def get_work_properties(topo, indices=None):
        """Default : no work vector
        """
        return DifferentialOperation._find_work(0, topo, indices)

    @staticmethod
    def _find_work(lwork, topo, indices=None):
        """Internal function to find the shape of work vector
        for a given topo and a set of indices
        """
        if lwork == 0:
            return {'rwork': None, 'iwork': None}

        if indices is not None:
            if not Utils.is_on_proc(indices):
                shape = (0, ) * topo.domain.dim
            else:
                shape = np.prod(tuple([indices[i].stop - indices[i].start
                                       for i in xrange(len(indices))]))
        else:
            shape = np.prod(topo.mesh.local_resolution)
        return {'rwork': [(shape, ), ] * lwork, 'iwork': None}

    def _init_fd_method(self, topo, reduce_output_shape):
        """Build the finite difference scheme
        """
        msg = 'FD scheme Not yet implemented for this operation.'
        assert self.method.__mro__[0] in self._authorized_methods, msg
        if not self._on_proc:
            empty_set = [slice(0, 0), ] * self._dim
            return self.method(topo.mesh.space_step,
                               empty_set)

        fd_scheme = self.method(topo.mesh.space_step,
                                self.in_indices,
                                reduce_output_shape)
        msg = 'Ghost layer is too small for the chosen FD scheme.'
        required_ghost_layer = fd_scheme.ghosts_layer_size
        assert (topo.ghosts() >= required_ghost_layer).all(), msg
        return fd_scheme


class Curl(DifferentialOperation):
    """
    Computes nabla X V, V being a vector field.
    """
    _authorized_methods = [FDC2, FDC4]
    _lwork = 1

    def __init__(self, **kwds):
        """Curl of a vector field
        """
        super(Curl, self).__init__(**kwds)
        assert len(self.in_indices) > 1
        # connect to fd function call
        self.fcall = self._central
        if len(self.in_indices) == 2:
            # a 'fake' curl for 2D case
            self.fcall = self._central_2d

    @staticmethod
    def get_work_properties(topo, indices=None):
        return Curl._find_work(Curl._lwork, topo, indices)

    def __call__(self, variable, result):
        if self._on_proc:
            return self.fcall(variable, result)
        else:
            return result

    def _central(self, variable, result):
        """ 3D Curl

        Parameters
        ----------
        variable : list of numpy arrays
            the input vector field
        result : list of numpy arrays
            in/out result
        """
        assert len(result) == len(variable)
        #--  d/dy vz -- in result[XDIR]
        self.fd_scheme.compute(variable[ZDIR], YDIR, result[XDIR])
        # -- -d/dz vy -- in work
        self.fd_scheme.compute(variable[YDIR], ZDIR, self._work[0])
        # result_x = d/dy vz - d/dz vy
        result[XDIR][self.output_indices] -= self._work[0][self.output_indices]

        #--  d/dz vx -- in result[YDIR]
        self.fd_scheme.compute(variable[XDIR], ZDIR, result[YDIR])
        # -- -d/dx vz -- in work
        self.fd_scheme.compute(variable[ZDIR], XDIR, self._work[0])
        # result_y = d/dz vx - d/dx vz
        result[YDIR][self.output_indices] -= self._work[0][self.output_indices]

        #-- d/dx vy in result[ZDIR]
        self.fd_scheme.compute(variable[YDIR], XDIR, result[ZDIR])
        # result_z = d/dx vy - d/dy vx
        self.fd_scheme.compute(variable[XDIR], YDIR, self._work[0])
        result[ZDIR][self.output_indices] -= self._work[0][self.output_indices]

        return result

    def _central_2d(self, variable, result):
        """ 2D Curl
        Parameters
        ----------
        variable : list of numpy arrays
            the input vector field
        result : list of numpy arrays
            in/out result
        """
        assert len(result) == 1
        #-- d/dx vy in result[ZDIR]
        self.fd_scheme.compute(variable[YDIR], XDIR, result[0])
        # result_z = d/dx vy - d/dy vx
        self.fd_scheme.compute(variable[XDIR], YDIR, self._work[0])
        result[0][self.output_indices] -= self._work[0][self.output_indices]
        return result


class DivRhoV(DifferentialOperation):
    """
    Computes \f$ \nabla.(\rho V) \f$, \f$ \rho\f$ a scalar field and
    V a vector field.
    Works for any dimension.

    Methods : FDC4
    """

    _authorized_methods = [FDC4]

    def __init__(self, **kwds):
        """Divergence of rho V, rho a scalar field, V a vector field

        Parameters
        ----------
        fd_optim : bool, optional
            if 'CAA', compute result += div(rhoV) else
            result = div(rhoV), which is the default.
        **kwds : parameters for base class

        Default method : FDC4
        """
        self._in_gh = None
        self._wk_indices = None
        self._fd_work = None
        super(DivRhoV, self).__init__(**kwds)

    def _init_fd_method(self, topo, reduce_output_shape):
        """Build the finite difference scheme
        """
        msg = 'FD scheme Not yet implemented for this operation.'
        assert self.method.__mro__[0] in self._authorized_methods, msg
        # for this operation, the fd scheme uses
        # work array as input, which means that work
        # array must include ghost points. This is done
        # thanks to self.wk_indices
        if not self._on_proc:
            empty_set = [slice(0, 0), ] * self._dim
            self._fd_work = self.method(topo.mesh.space_step,
                                        empty_set)
            self._in_gh = empty_set
            self._wk_indices = empty_set
            return self._fd_work

        gh = self.method.ghosts_layer_size
        ref = self.in_indices
        self._in_gh = [slice(ref[i].start - gh, ref[i].stop + gh)
                       for i in xrange(self._dim)]
        for sl in self._in_gh:
            assert sl.start >= 0
            assert sl.stop >= 0
        wk_shape = tuple([self._in_gh[i].stop - self._in_gh[i].start
                          for i in xrange(self._dim)])
        self._wk_indices = [slice(gh, wk_shape[i] - gh)
                            for i in xrange(self._dim)]
        if reduce_output_shape:
            iout = [slice(0, wk_shape[i] - 2 * gh)
                    for i in xrange(self._dim)]
        else:
            iout = None
        fd_scheme = self.method(topo.mesh.space_step,
                                self._wk_indices,
                                output_indices=iout)
        self._fd_work = self.method(topo.mesh.space_step,
                                    self._wk_indices)
        msg = 'Ghost layer is too small for the chosen FD scheme.'
        required_ghost_layer = fd_scheme.ghosts_layer_size
        assert (topo.ghosts() >= required_ghost_layer).all(), msg
        return fd_scheme

    def _set_work_arrays(self, topo, reduce_output_shape, rwork=None):
        """Check and allocate internal work buffers.
        """
        wk_prop = self.get_work_properties(topo, self.in_indices)['rwork']
        if wk_prop is None:
            return []
        subshape = np.asarray([self.in_indices[i].stop -
                               self.in_indices[i].start
                               for i in xrange(self._dim)])
        if self._on_proc:
            subshape += 2 * self.fd_scheme.ghosts_layer_size
        subshape = tuple(subshape)

        return WorkSpaceTools.check_work_array(len(wk_prop), subshape, rwork)

    @staticmethod
    def get_work_properties(topo, indices=None):
        # fd scheme will be applied on work so it needs ghost points.
        if indices is not None:
            if not Utils.is_on_proc(indices):
                shape = (0, ) * topo.domain.dim
            else:
                shape = np.asarray([indices[i].stop - indices[i].start
                                    for i in xrange(len(indices))])
                shape += 2 * topo.ghosts()
        else:
            shape = np.asarray(topo.mesh.local_resolution).copy()
        shape = np.prod(shape)
        return {'rwork': [(shape,), ] * 2, 'iwork': None}

    def __call__(self, var1, scal, result):
        """Apply operation

        Parameters
        ----------
        var1 : list of numpy arrays
            the vector field 'V'
        scal : list of numpy array
            the scalar field
        result : list of numpy arrays
            in/out buffer
        Returns
        -------
        numpy array

        """
        assert scal is not result
        assert len(result) == len(scal)
        for i in xrange(len(var1)):
            assert var1[i] is not result
        if self._on_proc:
            return self._central4(var1, scal, result)
        else:
            return result

    def _central4(self, var1, scal, result):
        """
        Compute central finite difference scheme, order 4.
        No work vector provided by user --> self._work will be
        used. It must be created at init and thus memshape is required.
        """

        # _work[0:1] are used as temporary space
        # for computation
        # div computations are accumulated into result.
        # Result does not need initialisation to zero.

        # d/dx (scal * var1x), saved into result
        np.multiply(scal[0][self._in_gh],
                    var1[XDIR][self._in_gh], self._work[0])
        result[0] = self.fd_scheme.compute(self._work[0], XDIR, result[0])
        # other components (if any) saved into _work[0] and added into result
        # d/dy (scal * var1y), saved in work and added into result
        # d/dz(scal * var1z), saved in work and added into result (if 3D)
        for cdir in xrange(1, len(var1)):
            np.multiply(scal[0][self._in_gh],
                        var1[cdir][self._in_gh], self._work[0])
            self._work[1] = self._fd_work.compute(self._work[0], cdir,
                                                  self._work[1])
            result[0][self.output_indices] += self._work[1][self._wk_indices]

        return result


class DivWV(DivRhoV):
    """
    Computes nabla.(W.Vx, W.Vy, W.Vz), W and V some vector fields.

    """

    _authorized_methods = [FDC4]

    def __init__(self, **kwds):
        """Divergence of (W.Vx, W.Vy, W.Vz), W, V two vector fields

        Parameters
        ----------
        **kwds : parameters for base class

        Default method : FDC4
        """
        super(DivWV, self).__init__(**kwds)
        msg = 'Implemented only in 3D.'
        assert self._dim == 3, msg

    def __call__(self, var1, var2, result):
        """Apply operation

        Parameters
        ----------
        var1 : list of numpy arrays
            the vector field 'W'
        var2 : list of numpy arrays
            the vector field 'V'
        result : list of numpy arrays
            in/out buffers
        Returns
        -------
        list of numpy arrays

        """
        assert len(result) == len(var2)
        if self._on_proc:
            for cdir in xrange(len(var2)):
                result[cdir:cdir + 1] = self._central4(
                    var1, var2[cdir:cdir + 1], result=result[cdir:cdir + 1])

        return result


class Laplacian(DifferentialOperation):
    """Computes  the laplacian of a field.
    """
    _authorized_methods = [FD2C2]
    _lwork = 1

    @staticmethod
    def get_work_properties(topo, indices=None):
        return Laplacian._find_work(Laplacian._lwork, topo, indices)

    def __call__(self, var, result):
        """"Apply laplacian

        Parameters
        ----------
        var : list of numpy arrays
            the scalar field
        result : list of numpy arrays
            in/out buffers
        Returns
        -------
        numpy array
        """
        assert len(var) == len(result)
        if self._on_proc:
            for d in xrange(len(var)):
                self.fd_scheme.compute(var[d], 0, result[d])
                for cdir in xrange(1, self._dim):
                    self.fd_scheme.compute_and_add(var[d], cdir, result[d],
                                                   self._work[0])
        return result


class GradS(DifferentialOperation):
    """Gradient of a scalar field
    """
    _authorized_methods = [FDC4, FDC2]

    def __call__(self, scal, result):
        """Apply gradient, with central finite difference scheme.

        Parameters
        ----------
        scal : list of numpy arrays
            the input scalar field
        result : list of numpy arrays
            in/out result
        """
        assert len(result) == self._dim
        if self._on_proc:
            for cdir in xrange(self._dim):
                #  d/dcdir (scal), saved in data[cdir]
                self.fd_scheme.compute(scal[0], cdir, result[cdir])

        return result


class GradV(DifferentialOperation):
    """Gradient of a vector field
    """
    _authorized_methods = [FDC4, FDC2]

    def __call__(self, var1, result):
        """Apply gradient

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : list of numpy arrays
            in/out result
        """
        nbc = len(var1)
        assert len(result) == nbc * self._dim
        if self._on_proc:
            for cdir in xrange(self._dim):
                pos = nbc * cdir
                for ddir in xrange(self._dim):
                    #  d/dddir (v[cdir]), saved in result[pos]
                    self.fd_scheme.compute(var1[cdir], ddir, result[pos])
                    pos += 1

        return result


class GradVxW(DifferentialOperation):
    """Computes [nabla(V)][W] with
    V and W some vector fields.
    """
    _authorized_methods = [FDC4, FDC2]
    _lwork = 2

    @staticmethod
    def get_work_properties(topo, indices=None):
        return DifferentialOperation._find_work(GradVxW._lwork, topo, indices)

    def __call__(self, var1, var2, result, diagnostics):
        """Apply gradient, with central finite difference scheme.

        Parameters
        ----------
        var1, var2 : list of numpy arrays
           the input vector fields
        result : list of numpy arrays
            in/out result. Overwritten.
        diagnostics : numpy array
            some internal diagnostics (like max of div(v) ...).
            In/out param, overwritten
        """
        assert len(result) == len(var1)
        if self._on_proc:
            diagnostics[:] = 0.
            nbc = len(var1)
            for comp in xrange(nbc):
                result[comp][...] = 0.0
                self._work[1][...] = 0.0
                for cdir in xrange(self._dim):
                    # self._work = d/dcdir (var1_comp)
                    self.fd_scheme.compute(var1[comp], cdir, self._work[0])
                    # some diagnostics ...
                    if cdir == comp:
                        tmp = np.max(abs(self._work[0][self.output_indices]))
                        diagnostics[0] = max(diagnostics[0], tmp)
                    np.add(abs(self._work[0]), self._work[1], self._work[1])

                    # compute self._work = self._work.var2[cdir]
                    np.multiply(self._work[0][self.output_indices],
                                var2[cdir][self.in_indices],
                                self._work[0][self.output_indices])
                    # sum to obtain nabla(var_comp) . var2,
                    # saved into result[comp]
                    npw.add(result[comp][self.output_indices],
                            self._work[0][self.output_indices],
                            result[comp][self.output_indices])
                diagnostics[1] = max(
                    diagnostics[1], np.max(self._work[1][self.output_indices]))

        return result, diagnostics


class DivAdvection(DifferentialOperation):
    """ Computes -nabla .(V . nabla V) with V a vector field.
    """
    _authorized_methods = [FDC4, FDC2]
    _lwork = 3

    @staticmethod
    def get_work_properties(topo, indices=None):
        return DifferentialOperation._find_work(DivAdvection._lwork,
                                                topo, indices)

    def __call__(self, var1, result):
        """Apply divergence, with central finite difference scheme.

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : list of numpy array
            in/out result, overwritten.
        """
        assert len(result) == 1
        nbc = len(var1)
        if not self._on_proc:
            return result
        assert nbc == 3, 'Only 3D case is implemented.'
        # Compute diff(var[dir], dir), saved in work[dir]
        for d in xrange(nbc):
            self.fd_scheme.compute(var1[d], d, self._work[d])

        # result = Vx,x * Vy,y
        np.multiply(self._work[XDIR][self.output_indices],
                    self._work[YDIR][self.output_indices],
                    result[0][self.output_indices])
        # wk[0] = Vx,x * Vz,z
        np.multiply(self._work[XDIR][self.output_indices],
                    self._work[ZDIR][self.output_indices],
                    self._work[XDIR][self.output_indices])
        # result = result + Vx,x * Vz,z
        np.add(self._work[XDIR][self.output_indices],
               result[0][self.output_indices],
               result[0][self.output_indices])
        # wk[1] = Vy,y * Vz,z
        np.multiply(self._work[YDIR][self.output_indices],
                    self._work[ZDIR][self.output_indices],
                    self._work[YDIR][self.output_indices])
        # result = result + Vy,y * Vz,z
        np.add(self._work[YDIR][self.output_indices],
               result[0][self.output_indices],
               result[0][self.output_indices])

        self.fd_scheme.compute(var1[XDIR], YDIR, self._work[0])
        self.fd_scheme.compute(var1[YDIR], XDIR, self._work[1])
        # wk[0] = Vx,y * Vy,x
        np.multiply(self._work[0][self.output_indices],
                    self._work[1][self.output_indices],
                    self._work[0][self.output_indices])
        # result = result - Vx,y * Vy,x
        np.subtract(result[0], self._work[0], result[0])

        self.fd_scheme.compute(var1[XDIR], ZDIR, self._work[0])
        self.fd_scheme.compute(var1[ZDIR], XDIR, self._work[1])
        # wk[0] = Vx,z * Vz,x
        np.multiply(self._work[0], self._work[1], self._work[0])
        # result = result - Vx,z * Vz,x
        np.subtract(result[0][self.output_indices],
                    self._work[0][self.output_indices],
                    result[0][self.output_indices])
        self.fd_scheme.compute(var1[YDIR], ZDIR, self._work[0])
        self.fd_scheme.compute(var1[ZDIR], YDIR, self._work[1])
        # wk[0] = Vy,z * Vz,y
        np.multiply(self._work[0][self.output_indices],
                    self._work[1][self.output_indices],
                    self._work[0][self.output_indices])
        # result = result - Vy,z * Vz,y
        np.subtract(result[0][self.output_indices],
                    self._work[0][self.output_indices],
                    result[0][self.output_indices])

        result[0][self.output_indices] *= 2.0
        return result


class Strain(DifferentialOperation):
    """Compute 0.5(grad U + grad U^T), U a vector field
    """
    _authorized_methods = [FDC4, FDC2]

    def __init__(self, **kwds):
        super(Strain, self).__init__(**kwds)
        if self._dim == 3:
            self._i1 = [0, 0, 1]
            self._i2 = [1, 2, 2]
        elif self._dim == 2:
            self._i1 = [0, ]
            self._i2 = [1, ]

    def __call__(self, var1, result):
        """Compute strain

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : list of numpy arrays
            in/out result

        result = [diagonal components, extra-diag comp]

        For example in 3D :

        result = [Strain00, Strain11, Strain22, Strain01, Strain02, Strain12]
        """
        nbc = len(var1)
        assert len(result) == self._dim * (self._dim + 1) / 2
        if not self._on_proc:
            return result

        pos = nbc
        # Extra diagonal terms. result[0] is used as temporary
        # work array.
        for i1, i2 in zip(self._i1, self._i2):
            self.fd_scheme.compute(var1[i1], i2, result[pos])
            self.fd_scheme.compute_and_add(var1[i2], i1, result[pos],
                                           result[0])
            np.multiply(result[pos], 0.5, result[pos])
            pos += 1
        # Then diagonal terms
        for cdir in xrange(self._dim):
            self.fd_scheme.compute(var1[cdir], cdir, result[cdir])

        return result


class StrainCriteria(DifferentialOperation):
    """For Strain = 0.5(grad U + grad U^T), U a vector field
    compute max(sum_j abs(Strain_i,j))

    Warning : this is a local (mpi) max computation, since input array var1
    is also a local array.
    """
    _authorized_methods = [FDC4, FDC2]

    def __init__(self, **kwds):
        super(StrainCriteria, self).__init__(**kwds)
        if self._dim == 3:
            self._i1 = [0, 0, 1]
            self._i2 = [1, 2, 2]
            self._pos = [0, 1, 2]
            self._adds = [[0, 1], [0, 2], [1, 2]]
        elif self._dim == 2:
            self._i1 = [0, ]
            self._i2 = [1, ]
            self._pos = [0]
            self._adds = [[0, ], [0, ]]

    @staticmethod
    def get_work_properties(topo, indices=None):
        dime = topo.domain.dim
        if dime == 2:
            lwork = 2
        elif dime == 3:
            lwork = 4

        return DifferentialOperation._find_work(lwork, topo, indices)

    def __call__(self, var1, result):
        """Compute criteria

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : numpy array
            criteria values for each direction.

        """
        work = self._work
        assert len(result) == self._dim
        if not self._on_proc:
            return result

        # First, compute extra-diagonal terms
        for i1, i2, ipos in zip(self._i1, self._i2, self._pos):
            self.fd_scheme.compute(var1[i1], i2, work[ipos])
            self.fd_scheme.compute_and_add(var1[i2], i1, work[ipos],
                                           work[-1])
            np.multiply(work[ipos], 0.5, work[ipos])
            np.abs(work[ipos], work[ipos])

        # Then add those terms to diagonal term and compute max
        for idiag in xrange(self._dim):
            self.fd_scheme.compute(var1[idiag], idiag, work[-1])
            np.abs(work[-1], work[-1])
            for i1 in self._adds[idiag]:
                np.add(work[i1], work[-1], work[-1])
            if self._on_proc:
                result[idiag] = np.max(work[-1][self.output_indices])
            else:
                result[idiag] = 0.0
        return result


class MaxDiagGradV(DifferentialOperation):
    """Maximum of each diagonal term (abs) of the Gradient of a vector field
    """
    _authorized_methods = [FDC4, FDC2]
    _lwork = 1

    @staticmethod
    def get_work_properties(topo, indices=None):
        return DifferentialOperation._find_work(MaxDiagGradV._lwork,
                                                topo, indices)

    def __call__(self, var1, result):
        """Compute max

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : numpy array
            in/out result
        """
        assert len(result) == self._dim
        if not self._on_proc:
            return result

        for cdir in xrange(self._dim):
            #  d/dcdir (var1[cdir]) saved in work
            self._work[0] = self.fd_scheme.compute(var1[cdir],
                                                   cdir, self._work[0])
            np.abs(self._work[0], self._work[0])
            if self._on_proc:
                result[cdir] = np.max(self._work[0][self.output_indices])

        return result


class StretchLike(DifferentialOperation):
    """Maximum of the sum of each 'line' (abs) of the Gradient of a vector field

    Warning: local (mpi) maximum.
    """
    _authorized_methods = [FDC4, FDC2]
    _lwork = 2

    @staticmethod
    def get_work_properties(topo, indices=None):
        return DifferentialOperation._find_work(StretchLike._lwork,
                                                topo, indices)

    def __call__(self, var1, result):
        """Compute max

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : numpy array
            in/out result

        result[i] = max(sum_j abs(diff(u_i, j)) for i = 0..dim
        """
        assert len(result) == self._dim
        result[...] = 0.
        if not self._on_proc:
            return result
        for cdir in xrange(self._dim):
            self._work[1][...] = 0.0
            for ddir in xrange(self._dim):
                #  d/dddir (var1[cdir]) saved in work
                self._work[0] = self.fd_scheme.compute(var1[cdir],
                                                       ddir, self._work[0])
                np.abs(self._work[0], self._work[0])
                np.add(self._work[0], self._work[1], self._work[1])
            if self._on_proc:
                result[cdir] = np.max(self._work[1][self.output_indices])

        return result


class DiagAndStretch(DifferentialOperation):
    """Maximum of the sum of each 'line' of abs of the Gradient of a vector field
    and max of each abs of 'diagonal' block of the gradient.

    Warning : this is a local (mpi) max computation, since input array var1
    is also a local array.
    """
    _authorized_methods = [FDC4, FDC2]
    _lwork = 2

    @staticmethod
    def get_work_properties(topo, indices=None):
        return DifferentialOperation._find_work(DiagAndStretch._lwork,
                                                topo, indices)

    def __call__(self, var1, result):
        """Compute max

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : numpy array
            in/out result

        result[i] = max(abs(diff(u_i,i))) for i = 0..dim
        result[dim:...] = max(sum_j abs(diff(u_i, j)) for i = 0..dim
        """
        assert len(result) == 2 * self._dim
        result[...] = 0.

        if not self._on_proc:
            return result

        pos = 0
        for cdir in xrange(self._dim):
            self._work[1][...] = 0.0
            for ddir in xrange(self._dim):
                #  d/dcdir (var1[cdir]) saved in work
                self._work[0] = self.fd_scheme.compute(var1[cdir],
                                                       ddir, self._work[0])
                np.abs(self._work[0], self._work[0])
                if cdir == ddir and self._on_proc:
                    result[pos] = np.max(self._work[0][self.output_indices])
                    pos += 1
                np.add(self._work[0], self._work[1], self._work[1])
            if self._on_proc:
                result[self._dim + cdir] = np.max(
                    self._work[1][self.output_indices])

        return result


class StrainAndStretch(DifferentialOperation):
    """For Strain = 0.5(grad U + grad U^T), U a vector field
    compute max(sum_j abs(Strain_i,j)) and maximum of the sum
    of each 'line' (abs) of the Gradient of U.

    Warning : this is a local (mpi) max computation, since input array var1
    is also a local array.

    result = [ max(sum_j(|strain_0,j|), ...],
               max(sum_j(|u0,j|), max(sum_j(|u1,j|)], ...]
    """
    _authorized_methods = [FDC4, FDC2]

    def __init__(self, **kwds):
        super(StrainAndStretch, self).__init__(**kwds)
        if self._dim == 3:
            self._fcall = self._call_3d
        elif self._dim == 2:
            self._fcall = self._call_2d

    @staticmethod
    def get_work_properties(topo, indices=None):
        dime = topo.domain.dim
        if dime == 2:
            lwork = 4
        elif dime == 3:
            lwork = 6

        return DifferentialOperation._find_work(lwork, topo, indices)

    def __call__(self, var1, result):
        """Compute criteria

        Parameters
        ----------
        var1 : list of numpy arrays
            the input vector field
        result : numpy array
            criteria values for each direction.

        """
        if self._on_proc:
            return self._fcall(var1, result)
        else:
            return result

    def _call_3d(self, var1, result):
        work = self._work
        assert len(result) == 2 * self._dim
        # work[0] = abs(eps_x,x)
        self.fd_scheme.compute(var1[XDIR], XDIR, work[0])
        np.abs(work[0], work[0])

        # work[1] = diff(u(x), i2)
        # work[2] = abs(work[1])
        # work[3] = diff(u(x), i3)
        # work[4] = abs(work[3])
        self.fd_scheme.compute(var1[XDIR], YDIR, work[1])
        np.abs(work[1], work[2])
        self.fd_scheme.compute(var1[XDIR], ZDIR, work[3])
        np.abs(work[3], work[4])
        # max(sum_j(abs(diff(u(x),j))))
        np.add(work[2], work[4], work[2])
        np.add(work[2], work[0], work[2])
        result[3] = np.max(work[2][self.output_indices])
        # free: p[2, 4, 5]
        # work[5] = diff(u(y), x)
        self.fd_scheme.compute(var1[YDIR], XDIR, work[5])
        # ... used to compute work[1] = abs(eps_x,y)
        np.add(work[1], work[5], work[1])
        np.multiply(work[1], 0.5, work[1])
        np.abs(work[1], work[1])
        # work[4] = diff(u(z), x)
        self.fd_scheme.compute(var1[ZDIR], XDIR, work[4])
        # ... used to compute work[2] = abs(eps_x,z)
        np.add(work[4], work[3], work[2])
        np.multiply(work[2], 0.5, work[2])
        np.abs(work[2], work[2])
        # At this point,
        # work[p[0, 1, 2] = abs([epsxx, epsxy, epsxz])
        # compute max(epsxx + epsxy + epsxz)
        np.add(work[1], work[0], work[0])
        np.add(work[2], work[0], work[0])
        result[0] = np.max(work[0][self.output_indices])
        # work[4] = abs(diff(u(z), x))
        # work[5] = abs(diff(u(y), x))
        np.abs(work[4], work[4])
        np.abs(work[5], work[5])
        # free : p[0, 3]
        # work[0] = abs(eps_y,y)
        self.fd_scheme.compute(var1[YDIR], YDIR, work[0])
        np.abs(work[0], work[0])
        np.add(work[0], work[1], work[1])
        np.add(work[0], work[5], work[5])
        # work[0] = abs(diff(u(y), z))
        self.fd_scheme.compute(var1[YDIR], ZDIR, work[3])
        np.abs(work[3], work[0])
        np.add(work[0], work[5], work[5])
        result[4] = np.max(work[5][self.output_indices])
        # free : p[0, 5]
        # work[0] = diff(u(i1), i2)
        # work[5] = abs(diff(u(i1), i2))
        self.fd_scheme.compute(var1[ZDIR], YDIR, work[0])
        np.abs(work[0], work[5])
        # ... used to compute work[0] = abs(eps_yz)
        np.add(work[0], work[3], work[0])
        np.multiply(work[0], 0.5, work[0])
        np.abs(work[0], work[0])
        # compute max(epsyy + epsxy + epsyz)
        np.add(work[1], work[0], work[1])
        result[1] = np.max(work[1][self.output_indices])
        np.add(work[2], work[0], work[2])
        # work[0] = abs(diff(uzz)
        self.fd_scheme.compute(var1[2], 2, work[0])
        np.abs(work[0], work[0])
        np.add(work[2], work[0], work[2])
        result[2] = np.max(work[2][self.output_indices])
        np.add(work[0], work[4], work[0])
        np.add(work[0], work[5], work[0])
        result[5] = np.max(work[0][self.output_indices])
        # ouf!

        return result

    def _call_2d(self, var1, result):
        work = self._work
        assert len(result) == 2 * self._dim
        # work[0] = abs(eps_x,x)
        self.fd_scheme.compute(var1[XDIR], XDIR, work[0])
        np.abs(work[0], work[0])
        
        # work[1] = diff(u(x), y)
        # work[2] = abs(diff(u(x), y))
        # work[3] = diff(u(y), x)
        self.fd_scheme.compute(var1[XDIR], YDIR, work[1])
        np.abs(work[1], work[2])
        self.fd_scheme.compute(var1[YDIR], XDIR, work[3])
        np.add(work[0], work[2], work[2])
        result[2] = np.max(work[2][self.output_indices])
        # work[1] = epsxy
        np.add(work[1], work[3], work[1])
        np.multiply(work[1], 0.5, work[1])
        np.abs(work[1], work[1])
        np.add(work[0], work[1], work[0])
        result[0] = np.max(work[0][self.output_indices])
        # work[0] = abs(eps_y,y)
        self.fd_scheme.compute(var1[YDIR], YDIR, work[0])
        np.abs(work[0], work[0])
        # work[2] = abs(uy,x)
        np.abs(work[3], work[2])
        np.add(work[2], work[0], work[2])
        result[3] = np.max(work[2][self.output_indices])
        np.add(work[0], work[1], work[0])
        result[0] = np.max(work[0][self.output_indices])
        return result
