"""ODE integrators

.. currentmodule hysop.numerics

* :class:`~odesolvers.Euler`,
* :class:`~odesolvers.RK2`,
* :class:`~odesolvers.RK3`,
* :class:`~odesolvers.RK4`,
* :class:`~odesolvers.ODESolver` (abstract base class).

See :ref:`odesolvers`

"""

from abc import ABCMeta, abstractmethod
from hysop.constants import WITH_GUESS
from hysop.numerics.update_ghosts import UpdateGhosts
from hysop.tools.misc import WorkSpaceTools
from hysop.numerics.odesolvers.runge_kutta import ExplicitRungeKutta

from hysop.tools.numpywrappers import npw
import numpy as np


class ODESolver(object):
    """Abstract base class for ODE solvers.

    To solve:

    dy(t)/dt = f(t,y)

    """
    __metaclass__ = ABCMeta

    def __init__(self, nb_components, topo, f=None, indices=None,
                 rwork=None, optim=None, wk_shape=None, need_synchro=False):
        """
        Parameters
        ----------
        nb_components : int
            number of components of the right-hand side.
        topo : :class:`~hysop.topology.topology.CartesianTopology`
            mpi process and data distribution
        f : python function, optional.
            right hand side of the equation to solve.
            If None, f = 0, else f must have the following signature:
            res = f(t, y, f_work)
            - y, work and res must be lists of numpy arrays
        indices : list of slices, optional
            Represents the local mesh on which the operation
            will be applied,
            like compute_index in :class:`~hysop.domain.mesh.Mesh`.
            See details in notes.
        reduce_output_shape : boolean, optional
            True to return the result in a reduced array. See notes below.
        rwork : list of numpy arrays, optional
            temporary work space.
        optim : int, optional
            the level of optimization (memory management).
            Default = None.
        wk_shape: tuple, optional
            shape of the internal work vector. Default = input topo
            resolution.
        need_synchro: boolean, optional
            Set to true if y parameter of the rhs needs an update
            of its ghost points before each call. This might be
            the case if rhs is an interpolation. Default=False

        Notes
        -----

        about optim parameter - Two levels are available:
        * default 'basic' level : at each, the rhs is evaluated and then the
        system is solved. This is the default behavior
        * optim = WITH_GUESS : in that case, work must contain a first
        evaluation of f(t,y) before each call.
        * wk_shape might be usefull when the advected field shape
        in advection dir is lower than topo shape (for instance when
        pushing particles with a threshhold parameter).

        """
        # RHS.
        self.f = f
        if f is None:
            self.f = lambda t, y, work: [npw.zeros_like(y[i])
                                         for i in xrange(nb_components)]
        if optim is None:
            self._fcall = self._basic
        elif optim is WITH_GUESS:
            self._fcall = self._core

        self._nb_components = nb_components

        if indices is None:
            indices = topo.mesh.compute_index
        self.in_indices = indices
        # True if work has to be done for this set
        # of indices on the current proc.
        #self._on_proc = Utils.is_on_proc(indices)
        #self.output_indices = None

        # work arrays
        self.rwork = self._set_work_arrays(topo, wk_shape, rwork=rwork)
        if need_synchro:
            self._synchronize = UpdateGhosts(topo, self._nb_components)
            self.compute_rhs = self._update_rhs_with_synchro
        else:
            self.compute_rhs = self.f

    def _set_work_arrays(self, topo, wk_shape=False, rwork=None):
        """Check and allocate internal work buffers.
        """
        wk_prop = self.get_work_properties(self._nb_components, topo)['rwork']
        if wk_prop is None:
            return []
        if wk_shape:
            subsize = wk_shape
        else:
            subsize = topo.mesh.local_resolution

        return WorkSpaceTools.check_work_array(len(wk_prop), subsize, rwork)

    @staticmethod
    def get_work_properties(nb_components, topo):
        """Returns the number of required internal work arrays for this method.

        Parameters
        ----------
        nb_components: integer,
            number of components of the right-hand-side
        topo : :class:`hysop.topology.topology.CartesianTopology`
            topology on which the integrator will be applied.
        """

    # def update_rhs_local(self, t, y, result):
    #     """Return rhs value, no synchronization between mpi domains.
    #     """
    #     return self.f(t, y, result)

    def _update_rhs_with_synchro(self, t, y, result):
        """"Return rhs value, but first
        synchronize y values between mpi domains
        (i.e. ghost points update)
        """
        self._synchronize(y)
        return self.f(t, y, result)

    def __call__(self, t, y, dt, result):
        """Apply integrator

        Parameters
        ----------
        t : double
            current time.
        y : list of numpy arrays
            vector field at time t.
        dt : double
            time step
        result : list of numpy arrays
            vector field at t + dt
        Returns
        -------
        result

        """
        return self._fcall(t, y, dt, result)

    def _basic(self, t, y, dt, result):
        """Apply integrator, with basic (i.e. no) memory
        optimisation.

        Parameters
        ----------
        t : double
            current time.
        y : list of numpy arrays
            vector field at time t.
        dt : double
            time step
        result : list of numpy arrays
            vector field at t + dt
        Returns
        -------
        result

        self.rwork is used as temp. array to compute the rhs.
        result provided as input arg.
        result may be equal to y.
        """
        self.rwork[:self._nb_components] = \
            self.compute_rhs(t, y, self.rwork[:self._nb_components])
        return self._core(t, y, dt, result)

    @abstractmethod
    def _core(self, t, y, dt, result):
        """internal function to integrate the system.

        Must not be called directly. All work, arrays ... must be properly
        allocated before any called to this function.

        Parameters
        ----------
        t : double
            current time.
        y : list of numpy arrays
            vector field at time t.
        dt : double
            time step
        result : list of numpy arrays
            vector field at t + dt
        Returns
        -------
        result

        """


class Euler(ODESolver):
    """Solve ode with forward Euler Method.

    """

    @staticmethod
    def get_work_properties(nb_components, topo):
        """Returns the number of required internal work arrays for this method.

        Parameters
        ----------
        nb_components: integer,
            number of components of the right-hand-side
        topo : :class:`hysop.topology.topology.CartesianTopology`
            topology on which the integrator will be applied.

        """
        shape = (np.prod(topo.mesh.local_resolution),)
        return {'rwork': [shape, ] * nb_components, 'iwork': None}

    def _core(self, t, y, dt, result):
        """Euler method computational core

        Must not be called directly. All work, arrays ... must be properly
        allocated before any called to this function.

        Parameters
        ----------
        t : double
            current time.
        y : list of numpy arrays
            vector field at time t.
        dt : double
            time step
        result : list of numpy arrays
            vector field at t + dt
        Returns
        -------
        result

        Highest level of optimization : result and self.rwork
        must be provided and work must contain a first evaluation of
        f(t,y)
        optim = WITH_GUESS
        """
        assert len(result) == self._nb_components
        # result = f(t, y)
        # result = y + dt * work0

        # result = y + work0
        for i in xrange(self._nb_components):
            np.add(y[i], self.rwork[i] * dt, result[i])
        return result

    @staticmethod
    def stability_coeff():
        """Returns a stability coefficient (used in stretching)
        """
        return 2.0


class RK2(ODESolver):
    """Solve ode with forward Runge-Kutta 2 method.
    """

    @staticmethod
    def get_work_properties(nb_components, topo):
        """Returns the number of required internal work arrays for this method.

        Parameters
        ----------
        nb_components: integer,
            number of components of the right-hand-side
        topo : :class:`hysop.topology.topology.CartesianTopology`
            topology on which the integrator will be applied.

        """
        shape = (np.prod(topo.mesh.local_resolution), )
        return {'rwork': [shape, ] * 2 * nb_components, 'iwork': None}

    def _core(self, t, y, dt, result):
        """RK2 computational core

        Must not be called directly. All work, arrays ... must be properly
        allocated before any called to this function.

        Parameters
        ----------
        t : double
            current time.
        y : list of numpy arrays
            vector field at time t.
        dt : double
            time step
        result : list of numpy arrays
            vector field at t + dt
        Returns
        -------
        result

        self.rwork[:nb_components] must contain a first evaluation of
        f(t, y)
        optim = WITH_GUESS

        Note : since result may be equal to y, it can not
        be used as a temporary workspace.
        """
        i_y = self.in_indices
        for i in xrange(self._nb_components):
            for j in xrange(len(self.rwork)):
                assert not npw.arrays_share_data(y[i], self.rwork[j])
        assert len(result) == self._nb_components
        work0 = self.rwork[:self._nb_components]
        yn = self.rwork[self._nb_components:2 * self._nb_components]

        # k1 = f(t,y) = work0
        # k2 = f(t + dt/2, y + dt/2 * k1)
        # result = y + dt * k2

        # yn
        #print ic, y[0].shape
        for i in xrange(self._nb_components):
            np.add(y[i][i_y], work0[i] * 0.5 * dt, yn[i])
            #np.add(y[i][ic], work0[i] * 0.5 * dt, yn[i])

        # k2 in work0
        work0 = self.compute_rhs(t + 0.5 * dt, yn, work0)
        # *= dt
        for i in xrange(self._nb_components):
            #np.multiply(work0[i], dt, work0[i])
            work0[i] *= dt
            # result = y + work0
            np.add(work0[i], y[i][i_y], result[i][i_y])
            #np.add(work0[i], y[i][ic], result[i])

        return result

    @staticmethod
    def stability_coeff():
        """Returns a stability coefficient (used in stretching)
        """
        return 2.0


class RK3(ODESolver):
    """ODESolver implementation for solving an equation system with RK3 method.
    """

    @staticmethod
    def get_work_properties(nb_components, topo):
        """Returns the number of required internal work arrays for this method.

        Parameters
        ----------
        nb_components: integer,
            number of components of the right-hand-side
        topo : :class:`hysop.topology.topology.CartesianTopology`
            topology on which the integrator will be applied.

        """
        shape = (np.prod(topo.mesh.local_resolution),)
        return {'rwork': [shape, ] * 3 * nb_components, 'iwork': None}

    def _core(self, t, y, dt, result):
        """RK3 computational core

        Must not be called directly. All work, arrays ... must be properly
        allocated before any called to this function.

        Parameters
        ----------
        t : double
            current time.
        y : list of numpy arrays
            vector field at time t.
        dt : double
            time step
        result : list of numpy arrays
            vector field at t + dt
        Returns
        -------
        result

        Highest level of optimization : work and result
        must be provided and
        work[:nb_components] must contain a first evaluation of
        f(t,y,f_work)
        optim = WITH_GUESS

        Note : since result may be equal to y, it can not
        be used as a temporary workspace.
        """
        for i in xrange(self._nb_components):
            cond = [y[i] is self.rwork[j] for j in xrange(len(self.rwork))]
            assert cond.count(True) is 0
        assert len(result) == self._nb_components

        work0 = self.rwork[:self._nb_components]
        yn = self.rwork[self._nb_components:2 * self._nb_components]
        kn = self.rwork[2 * self._nb_components:3 * self._nb_components]

        # k1 = f(t,y) = work0
        # k2 = f(t+dt/3, y + dt/3 *k1))
        # k3 = f(t + 2/3 *dt , y + 2/3 dt * k2))
        # result = y + 0.25 * dt * (k1 + 3 * k3)
        # yn
        for i in xrange(self._nb_components):
            np.add(y[i], work0[i] * dt / 3, yn[i])

        # k2 in kn
        kn = self.compute_rhs(t + dt / 3, yn, kn)
        # yn
        for i in xrange(self._nb_components):
            np.add(y[i], 2 * dt / 3 * kn[i], yn[i])

        # k3 in kn
        kn = self.compute_rhs(t + 2 * dt / 3, yn, kn)
        # k1 + 3 * k3 in work0
        for i in xrange(self._nb_components):
            np.add(work0[i], 3 * kn[i], work0[i])
            # *= dt / 4
            work0[i] *= 0.25 * dt
            # result = y + work0
            np.add(work0[i], y[i], result[i])

        return result

    @staticmethod
    def stability_coeff():
        """Returns a stability coefficient (used in stretching)
        """
        return 2.5127


class RK4(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK4 method.
    """
    @staticmethod
    def get_work_properties(nb_components, topo):
        """Returns the number of required internal work arrays for this method.

        nb_components: integer,
            number of components of the right-hand-side
        topo : :class:`hysop.topology.topology.CartesianTopology`
            topology on which the integrator will be applied.
        """
        shape = (np.prod(topo.mesh.local_resolution),)
        return {'rwork': [shape, ] * 3 * nb_components, 'iwork': None}

    _one_over_six = 1. / 6.

    def _core(self, t, y, dt, result):
        """RK4 computational core

        Must not be called directly. All work, arrays ... must be properly
        allocated before any called to this function.

        Parameters
        ----------
        t : double
            current time.
        y : list of numpy arrays
            vector field at time t.
        dt : double
            time step
        result : list of numpy arrays
            vector field at t + dt
        Returns
        -------
        result

        """
        for i in xrange(self._nb_components):
            cond = [y[i] is self.rwork[j] for j in xrange(len(self.rwork))]
            assert cond.count(True) is 0
        assert len(result) == self._nb_components

        work0 = self.rwork[:self._nb_components]
        yn = self.rwork[self._nb_components:2 * self._nb_components]
        kn = self.rwork[2 * self._nb_components:3 * self._nb_components]

        # k1 = f(t,y)
        # k2 = f(t + dt/2, y + dt/2 * k1)
        # k3 = f(t + dt/2, y + dt/2 * k2)
        # k4 = f(t + dt, y + dt * k3)
        # result = y + dt/6( k1 + 2 * k2 + 2 * k3 + k4)

        # yn
        for i in xrange(self._nb_components):
            np.add(y[i], work0[i] * dt / 2, yn[i])

        # k2 in kn
        kn = self.compute_rhs(t + dt / 2, yn, kn)

        # k1 + 2 * k2 in work0
        for i in xrange(self._nb_components):
            np.add(work0[i], 2 * kn[i], work0[i])
            # yn
            np.add(y[i], dt / 2 * kn[i], yn[i])

        # k3 in kn
        kn = self.compute_rhs(t + dt / 2, yn, kn)

        # k1 + 2 * k2 + 2 * k3 in work0
        for i in xrange(self._nb_components):
            np.add(work0[i], 2 * kn[i], work0[i])
            # yn
            np.add(y[i], dt * kn[i], yn[i])

        # K4 in kn
        kn = self.compute_rhs(t + dt, yn, kn)

        # k1 + 2 * k2 + 2 * k3 + k4
        for i in xrange(self._nb_components):
            np.add(work0[i], kn[i], work0[i])
            work0[i] *= self._one_over_six * dt
            # result = y + work0
            np.add(work0[i], y[i], result[i])

        return result

    @staticmethod
    def stability_coeff():
        """Returns a stability coefficient (used in stretching)
        """
        return 2.7853
