"""Tests for time integrators (RK ...)
"""
# -*- coding: utf-8 -*-
from hysop.numerics.odesolvers import Euler, RK2, RK3, RK4
from hysop.tools.numpywrappers import npw
import math
import numpy as np
from hysop.tools.parameters import Discretization
from hysop import Box
from hysop.tools.misc import WorkSpaceTools


pi = math.pi
sin = np.sin
cos = np.cos

# Grid resolution for tests
nx = ny = nz = 64
# Initial time
start = 0.
# Final time
end = 0.5

d1 = Discretization([nx + 1])
box = Box(length=[2.0 * pi], origin=[0.])
topo = box.create_topology(dim=1, discretization=d1)
local_res = topo.mesh.local_resolution
d3d = Discretization([nx + 1, ny + 1, nz + 1])
box3 = Box(length=[2.0 * pi, ] * 3, origin=[0., ] * 3)
topo3d = box3.create_topology(discretization=d3d)
local_res_3d = topo3d.mesh.local_resolution

# A set of tests and reference functions

def rhs(t, y, sol):
    sol[0][...] = -y[0]
    return sol


def rhs3d(t, y, work):
    work[0][...] = y[1]
    work[1][...] = -y[0]
    work[2][...] = -4 * y[0] * y[1]
    return work


def integrate(integ, nb_steps):
    """Execute odesolver 'integ' nb_steps times.

    * integ must be an instance of an odesolver (i.e. Euler, RK2 ...).
    * the system must be 1D.
    """
    t = start
    time_points = np.linspace(start, end, nb_steps)
    dtt = time_points[1] - time_points[0]
    y = [npw.ones(local_res[0]) * math.exp(-start)]
    res = [npw.zeros(local_res[0])]
    # work = None
    i = 1
    ref = npw.zeros((nb_steps, local_res[0]))
    ref[0, :] = y[0][:]
    while i < nb_steps:
        res = integ(t, y, dtt, res)
        y[0][...] = res[0]
        ref[i, :] = y[0][:]
        t += dtt
        i += 1
    err = 0.0
    for i in xrange(local_res[0]):
        err = max(err, (np.abs(ref[:, i] - np.exp(-time_points))).max())
    return dtt, err


def integrate_3d(integ, nb_steps):
    """Execute odesolver 'integ' nb_steps times.

    * integ must be an instance of an odesolver (i.e. Euler, RK2 ...).
    * 3D system
    """
    t = start
    time_points = np.linspace(start, end, nb_steps)
    dtt = time_points[1] - time_points[0]
    # y0
    y = [npw.zeros(local_res_3d), npw.ones(local_res_3d),
         npw.ones(local_res_3d)]
    res = [npw.zeros(local_res_3d), npw.zeros(local_res_3d),
           npw.zeros(local_res_3d)]
    # work = None
    i = 1
    xref = local_res_3d[0] // 3
    yref = local_res_3d[1] // 2
    zref = local_res_3d[2] // 5
    ref = npw.zeros((nb_steps, 4))
    ref[0, 1] = y[0][xref, yref, zref]
    ref[0, 2] = y[1][xref, yref, zref]
    ref[0, 3] = y[2][xref, yref, zref]
    ref[:, 0] = time_points
    while i < nb_steps:
        res = integ(t, y, dtt, res)
        for d in xrange(3):
            y[d][...] = res[d]
        ref[i, 1] = y[0][xref, yref, zref]
        ref[i, 2] = y[1][xref, yref, zref]
        ref[i, 3] = y[2][xref, yref, zref]
        t += dtt
        i += 1
    err = np.zeros(3)
    err[0] = (np.abs(ref[:, 1] - np.sin(time_points))).max()
    err[1] = (np.abs(ref[:, 2] - np.cos(time_points))).max()
    err[2] = (np.abs(ref[:, 3] - np.cos(2 * time_points))).max()
    return dtt, err


def run_integ(integrator, order):
    """apply 'integrator' (== Euler, RK3 ...)
    and check error according to input order. 1D case
    """
    nb_steps = 100
    wk_prop = integrator.get_work_properties(1, topo)['rwork']
    work = []
    for shape in wk_prop:
        work.append(npw.zeros(shape))
    dt, err = integrate(integrator(nb_components=1, topo=topo,
                                   f=rhs, rwork=work), nb_steps)
    assert err < dt ** order


def run_integ_3d(integrator, order):
    """apply 'integrator' (== Euler, RK3 ...)
    and check error according to input order. 3D case.

    """
    nb_steps = 50
    wk_prop = integrator.get_work_properties(3, topo3d)['rwork']
    work = []
    for shape in wk_prop:
        work.append(npw.zeros(shape))
    dt, err = integrate_3d(integrator(nb_components=3, topo=topo3d,
                                      f=rhs3d, rwork=work),
                           nb_steps)
    for e in err:
        assert e < dt ** order


def run_integ_3d_no_work(integrator, order):
    """apply 'integrator' (== Euler, RK3 ...)
    and check error according to input order. 3D case.
    """
    nb_steps = 50
    dt, err = integrate_3d(integrator(nb_components=3, topo=topo3d,
                                      f=rhs3d), nb_steps)
    for e in err:
        assert e < dt ** order


def test_euler_1d():
    run_integ(Euler, 1)


def test_rk2_1d():
    run_integ(RK2, 2)


def test_rk3_1d():
    run_integ(RK3, 3)


def test_rk4_1d():
    run_integ(RK4, 4)


def test_euler_3d():
    run_integ_3d(Euler, 1)
    run_integ_3d_no_work(Euler, 1)


def test_rk2_3d():
    run_integ_3d(RK2, 2)
    run_integ_3d_no_work(RK2, 2)


def test_rk3_3d():
    run_integ_3d(RK3, 3)
    run_integ_3d_no_work(RK3, 3)


def test_rk4_3d():
    run_integ_3d(RK4, 4)
    run_integ_3d_no_work(RK4, 4)


def test_user_defined_common_workspace():
    """One work space for all integrators"""
    integrators = {Euler: 1, RK2: 2, RK3: 3, RK4: 4}
    nb_steps = 50
    wk_prop = {}
    for integ in integrators:
        wk_prop[integ] = integ.get_work_properties(3, topo3d)['rwork']
    work = WorkSpaceTools.allocate_common_workspaces(wk_prop.values())
    for integ in integrators:
        dt, err = integrate_3d(integ(nb_components=3, topo=topo3d,
                                     f=rhs3d, rwork=work), nb_steps)
        for e in err:
            assert e < dt ** integrators[integ]
