# -*- coding: utf-8 -*-
"""Update time-step, depending on the flow field.

See :ref:`adaptive_time_step` for details.

"""
from hysop.constants import debug
from hysop.methods import TimeIntegrator, SpaceDiscretization
from hysop.operator.discrete.adapt_timestep import AdaptiveTimeStepD
from hysop.operator.continuous import opsetup
from hysop.operator.computational import Computational
import hysop.default_methods as default
from hysop.core.mpi import main_comm, MPI
from hysop.numerics.differential_operations import MaxDiagGradV, \
    DiagAndStretch, StrainCriteria, StretchLike, StrainAndStretch


class AdaptiveTimeStep(Computational):
    """The adaptative Time Step is computed according
    to the following expression :

    dt_adapt = min (dt_advection, dt_stretching, dt_cfl)

    """

    authorized_criteria = ['vort', 'gradU', 'stretch', 'cfl', 'strain']

    @debug
    def __init__(self, velocity, vorticity, simulation,
                 lcfl=0.125, cfl=0.5, time_range=None,
                 maxdt=9999., criteria=None, **kwds):
        """
        Parameters
        ----------
        velocity, vorticity : :class:`~hysop.fields.continuous_field.DiscreteFields`
            fields used to update the time steps. Read only.
        simulation : :class:`hysop.fields.problem.simulation.Simulation`
            time loop parameters description
        lcfl : double, optional
            the lagrangian CFL coefficient used for advection stability
        cfl : double, optional
            CFL coefficient
        criteria : list of strings, optional
            name of the criteria used to compute the new time step.
            See notes below.
        time_range : list of two integers, optional
            use to define a 'window' in which the current operator is applied.
            time_range = [start, end], outside this range, the operator
            has no effect. Start/end are iteration numbers.
            Default = [2, end of simu]
        maxdt : double, optional
            maximum value allowed for time step. Default = 9999.
        kwds : arguments passed to base class.

        Notes
        -----
       * This operator has no effect on input variables.
        * Authorized criteria are:
            * criteria for the advection scheme, one of :
                1. 'vort' (default)
                2. 'gradU'
                3. 'strain'
            * criteria for the stretching scheme :
                4. 'stretch'
            * cfl-like :
                5. 'cfl'
        * Computes a 'diagnostics' vector :
          diagnostics = (time, time_step, c1, c2, ...)
          ci = time step computed from the criterium number i in the list
          above.
        * dt_adapt = min(diagnostics, dtmax)
        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(AdaptiveTimeStep, self).__init__(variables=[velocity, vorticity],
                                               **kwds)
        if self.method is None:
            self.method = default.ADAPT_TIME_STEP
        assert SpaceDiscretization in self.method.keys()
        assert TimeIntegrator in self.method.keys()
        # Definition of criterion for dt_advec computation
        if criteria is None:
            criteria = ['vort']
        msg = 'criteria arg must be a list.'
        assert isinstance(criteria, list), msg
        msg = 'Unknow criteria : '
        for cr in criteria:
            assert cr in self.authorized_criteria, msg + cr
        msg = 'criteria list is empty!'
        assert len(criteria) >= 1, msg
        unknowns = [k for k in criteria if k not in self.authorized_criteria]
        msg = 'Unknown criteria ' + str(unknowns)
        assert len(unknowns) == 0, msg

        # list of criteria used to compute dt
        self.criteria = criteria

        # velocity variable (vector)
        self.velocity = velocity
        # vorticity variable (vector)
        self.vorticity = vorticity
        # simulation
        self.simulation = simulation

        self.input = self.variables
        self.output = []
        self.time_range = time_range
        self.lcfl, self.cfl, self.maxdt = lcfl, cfl, maxdt
        self._intercomms = {}
        self._set_inter_comm()

    def _set_inter_comm(self):
        """Create intercommunicators, if required (i.e. if there are several
        tasks defined in the domain).
        """
        task_is_source = self.mpi_params.task_id == self.domain.current_task()
        tasks_list = self.domain.tasks_list()
        others = (v for v in tasks_list if v != self.mpi_params.task_id)
        if task_is_source:
            remote_leader = set([tasks_list.index(i) for i in others])
        else:
            remote_leader = set([tasks_list.index(self.mpi_params.task_id)])

        for rk in remote_leader:
            self._intercomms[rk] = self.domain.task_comm.Create_intercomm(
                0, main_comm, rk)

    def get_work_properties(self):
        super(AdaptiveTimeStep, self).get_work_properties()
        diffop = None
        topo = self.discrete_fields[self.velocity].topology
        if 'gradU' in self.criteria:
            diffop = MaxDiagGradV
        if 'stretch' in self.criteria:
            diffop = StretchLike
        if 'strain' in self.criteria:
            diffop = StrainCriteria
        if 'stretch' in self.criteria and 'gradU' in self.criteria:
            diffop = DiagAndStretch
        if 'stretch' in self.criteria and 'strain' in self.criteria:
            diffop = StrainAndStretch
        if diffop is not None:
            return diffop.get_work_properties(topo)
        else:
            return {'rwork': None, 'iwork': None}

    def discretize(self):
        nb_ghosts = self.method[SpaceDiscretization].ghosts_layer_size
        super(AdaptiveTimeStep, self)._standard_discretize(nb_ghosts)

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        self.discrete_op =\
            AdaptiveTimeStepD(self.discrete_fields[self.velocity],
                              self.discrete_fields[self.vorticity],
                              self.simulation, method=self.method,
                              time_range=self.time_range,
                              criteria=self.criteria,
                              lcfl=self.lcfl,
                              cfl=self.cfl,
                              maxdt=self.maxdt,
                              rwork=rwork, iwork=iwork)
        # Output setup
        self._set_io('dt_adapt', (1, 7))
        self.discrete_op.set_writer(self._writer)
        self._is_uptodate = True

    def wait(self):
        task_is_source = self.mpi_params.task_id == self.domain.current_task()
        rank = self.mpi_params.rank
        dt = self.simulation.time_step
        for rk in self._intercomms:
            if task_is_source:
                # Local 0 broadcast current_indices to remote comm
                if rank == 0:
                    self._intercomms[rk].bcast(dt, root=MPI.ROOT)
                else:
                    self._intercomms[rk].bcast(dt, root=MPI.PROC_NULL)
            else:
                dt = self._intercomms[rk].bcast(dt, root=0)
                self.simulation.update_time_step(dt)

    def diagnostics(self):
        """Get the list of computed dt (for each criteria)
        """
        return self.discrete_op.diagnostics
