"""User-defined discrete operator"""

from hysop.operator.discrete.discrete import DiscreteOperator


class Custom(DiscreteOperator):
    """User-defined operator: action defined by an external function.
    """
    def __init__(self, function, in_fields, out_fields=None, **kwds):
        """

        Parameters
        ----------
        in_fields: list of :class:`~hysop.fields.discrete_field.DiscreteField`
             input fields args. for function, see notes below
        out_fields: list of :class:`~hysop.fields.discrete_field.DiscreteField`
             output fields args for function, see notes below
        function: python function
             a user defined function, called by this op.apply method.

        Notes
        -----
        A function is used to set the behavior of the current operator,
        during apply call.
        This function must look like::

            def some_func(simulation, in_fields, out_fields=None, diag=None):
                # do things ...

        and compute out_fields values.
        """
        # callback for apply function
        self.function = function
        super(Custom, self).__init__(**kwds)
        # in/out fields must obviously belong to variables
        self.input = in_fields
        if out_fields is not None:
            self.input += out_fields
            self.output = out_fields
        msg = 'Custom: all in/out fields must belong to op variables.'
        assert set(self.input).intersection(set(self.variables)) == \
            set(self.input), msg
        self._in_fields = in_fields
        self._out_fields = out_fields

    def apply(self, simulation=None):
        if self._writer is not None:
            diagnostics = self._writer.buffer
        else:
            diagnostics = None
        self.function(simulation, self._in_fields, self._out_fields,
                      diagnostics)
        ite = simulation.current_iteration
        if self._writer is not None and self._writer.do_write(ite):
            self._writer.write()
