# -*- coding: utf-8 -*-
"""Discrete Diffusion operator using FFTW (fortran)

See :ref:`diffusion` in HySoP user guide.

"""
try:
    from hysop.f2hysop import fftw2py
except ImportError:
    msg = 'fftw package not available for your hysop install.'
    msg += 'Try to recompile with WITH_FFTW=ON'
    raise ImportError(msg)

from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.constants import debug
from hysop.tools.profiler import profile


class DiffusionFFT(DiscreteOperator):
    """Discretized Poisson operator based on FFTW.
    See details in hysop.operator.diffusion.

    """
    @debug
    def __init__(self, viscosity, vorticity, **kwds):
        """Discrete diffusion operator, based on fftw solver.

        Parameters
        ----------
        viscosity : double
             constant viscosity value
        vorticity : :class:`~hysop.fields.discrete_field.DiscreteField`
             vorticity field, in/out parameter
        kwds : base class arguments
        """
        # Discretization of the solution field
        self.vorticity = vorticity
        # Viscosity.
        self.viscosity = viscosity
        dim = self.vorticity.dimension
        if dim == 3:
            self._apply = self.apply_3d
        elif dim == 2:
            self._apply = self.apply_2d
        else:
            raise AttributeError(dim + "D case not yet implemented.")
        # Base class initialisation
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(DiffusionFFT, self).__init__(variables=[vorticity],
                                           **kwds)
        self.input = [self.vorticity]
        self.output = [self.vorticity]

    @debug
    @profile
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing dt value for diffusion computation."
        dt = simulation.time_step
        ghosts = self.vorticity.topology.ghosts()
        self._apply(dt, ghosts)

    def apply_2d(self, dt, ghosts):
        """2d implementation of apply function"""
        self.vorticity.data = fftw2py.solve_diffusion_2d(
            self.viscosity * dt, self.vorticity.data, ghosts)

    def apply_3d(self, dt, ghosts):
        """3d implementation of apply function"""
        self.vorticity.data[0], self.vorticity.data[1],\
            self.vorticity.data[2] = \
            fftw2py.solve_diffusion_3d(self.viscosity * dt,
                                       self.vorticity.data[0],
                                       self.vorticity.data[1],
                                       self.vorticity.data[2],
                                       ghosts)

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        pass
        # TODO : fix bug that occurs when several finalize
        # of fft operators are called.
        # fftw2py.clean_fftw_solver(self.vorticity.dimension)


class CurlAndDiffusionFFT(DiscreteOperator):
    """Discretized Curl/Diffusion operator based on FFTW.
    See details in hysop.operator.diffusion.

    """
    @debug
    def __init__(self, viscosity, velocity, vorticity, **kwds):
        """Solve diffusion problem in Fourier
        domain (velocity curl + diffusion in one shot)

        Parameters
        ----------
        viscosity : double
             constant viscosity value
        velocity : :class:`~hysop.fields.discrete_field.DiscreteField`
             velocity field, input parameter
        vorticity : :class:`~hysop.fields.discrete_field.DiscreteField`
             velocity field, output parameter
        kwds : base class arguments (vorticity, viscosity ...)
        """
        self.velocity = velocity
        self.vorticity = vorticity
        self.viscosity = viscosity
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(CurlAndDiffusionFFT, self).__init__(
            variables=[velocity, vorticity], **kwds)
        self.input = [self.velocity]
        self.output = [self.vorticity]
        msge = 'CurlAndDiffusion: only implemented for 3d domain.'
        assert self.velocity.dimension == 3, msge

    @debug
    @profile
    def apply(self, simulation=None):
        dt = simulation.time_step
        gh_velo = self.velocity.topology.ghosts()
        gh_vorti = self.vorticity.topology.ghosts()
        self.vorticity.data[0], self.vorticity.data[1],\
            self.vorticity.data[2] = \
            fftw2py.solve_curl_diffusion_3d(self.viscosity * dt,
                                            self.velocity.data[0],
                                            self.velocity.data[1],
                                            self.velocity.data[2],
                                            self.vorticity.data[0],
                                            self.vorticity.data[1],
                                            self.vorticity.data[2],
                                            gh_velo, gh_vorti)
