"""
@file particle_advection.py
Base for directionally splitted advection solvers (pure-python and GPU version).
"""
from hysop.constants import debug, WITH_GUESS, HYSOP_REAL, HYSOP_INTEGER, SIZEOF_HYSOP_REAL, SIZEOF_HYSOP_INTEGER, ORDER
from hysop.methods import TimeIntegrator, Interpolation, Remesh, Support
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.backend.device.codegen.structs.mesh_info import MeshDirection
from hysop.tools.numpywrappers import npw
import hysop.default_methods as default
import numpy as np
from hysop.numerics.remeshing import Remeshing
from hysop.tools.profiler import profile

class ParticleAdvectionDir(DiscreteOperator):
    """
    Interface for particular advection method of a list of fields in a given direction.
    Base class for pure-python and GPU ParticleAdvectionDir implementations.
    """
    @debug
    def __init__(self, velocity, advected_fields, direction, **kwds):
        """
        Constructor.
        @param velocity:        discretization of the velocity field (all components)
        @param advected_fields : list of discretized fields to be advected.
        @param direction:      direction of advection
        """

        cls = self.__class__.__name__
        if ORDER != 'F':
            msg='{} has not been implemented for C-contiguous arrays!'.format(cls)
            raise NotImplementedError(msg)
        
        # Check input values
        if 'variables' in kwds.keys():
            msg='variables parameter is should not be used in {}.'.format(cls)
            raise ValueError(msg)
        if not isinstance(advected_fields, list):
            advected_fields = [advected_fields]
        if not isinstance(direction, int) or (direction not in MeshDirection.entries()):
            raise ValueError("Invalid direction value '{}'.".format(direction))
        if ('iwork' in kwds) or ('rwork' in kwds):
            msg='rwork or iwork parameters are useless in {}.'.format(cls)
            raise ValueError(msg)

        # Discrete fields and their topologies
        self.velocity        = velocity
        self.advected_fields = advected_fields

        self.velocity_topo   = velocity.topology
        self.fields_topo     = advected_fields[0].topology
        self._is_multi_scale = self._check_topologies()

        # MPI related variables
        self._comm           = self.fields_topo.comm
        self._comm_size      = self._comm.Get_size()
        self._comm_rank      = self._comm.Get_rank()
        self._is_distributed = (self._comm_size > 1)

        # Variables
        input_vars  = [velocity]+advected_fields
        output_vars = advected_fields
        variables   = input_vars

        super(ParticleAdvectionDir, self).__init__(
            rwork=None, iwork=None,
            variables=variables,
            **kwds)

        self.input     = self.variables
        self.output    = [df for df in self.variables if df is not self.velocity]
        self.direction = direction
    
    def _check_topologies(self):
        multiscale = False
        topo_ref = self.fields_topo
        for f in self.advected_fields:
            if f.topology != topo_ref:
                raise ValueError('One of the advected field has a different topology!')
        multiscale = (self.velocity_topo != topo_ref)
        return multiscale

# ParticleAdvectionDir interface    
    @staticmethod
    def supports_multiscale():
        return False
    @staticmethod
    def supports_mpi():
        return False
    
    def _check(self):
        cls = self.__class__.__name__
        if (self._is_multi_scale) and (not self.supports_multiscale()):
            msg='Multiscale advection in {} has not been implemented yet !'.format(cls)
            raise NotImplementedError(msg)
        if (self._is_distributed) and (not self.supports_mpi()):
            msg='MPI multi-GPU advection has not been implemented in {} yet!'.format(cls)
            raise NotImplementedError(msg)

    def _initialize(self):
        pass

    def get_work_properties(self):
        """ 
            directional work properties should return the same shapes
            to ease buffer sharing between all directions.
        """
        cls=self.__class__.__name__
        raise NotImplementedError('{}.get_work_properties'.format(cls))
    
    def setup(self, rwork, iwork):
        """ 
            rwork and iwork cannot be None during setup
    ,
    because they are shared between all directions.
        """
        cls=self.__class__.__name__
        raise NotImplementedError('{}.setup'.format(cls))

    def apply(self,simulation,step_id,**extra_params):
        """ 
            First check if input step_id corresponds to this operator's direction.
            And apply this operator with given step_id.
        """
        cls=self.__class__.__name__
        raise NotImplementedError('{}.apply'.format(cls))
    
    def step_directions(self):
        """ 
            List of directions in order of execution.
            
            Ex: If configured splitting method is Strang splitting of order 2
                which execution order is (X,Y,Z,Y,X) in 3d.
                this function should return 
                    [0]         in 1D
                    [0,1,0]     in 2D
                    [0,1,2,1,0] in 3D
        """
        cls=self.__class__.__name__
        raise NotImplementedError('{}.step_directions'.format(cls))



class PythonParticleAdvectionDir(ParticleAdvectionDir):
    """
    Pure python particle advection dir.
    """
    
    def __init__(self, velocity, advected_fields, direction, **kwds):
        super(self.__class__,self).__init__(velocity,advected_fields,direction,**kwds)
        
        #TODO check this implementation
        raise NotImplementedError('Check this implementation.')

        self._configure_numerical_methods()

    def _configure_numerical_methods(self):
        """
        Function to set the numerical method for python operator and link them
        to the proper working arrays.
        """
        # Use first field topology as reference
        topo = self.advected_fields[0].topology

        # --- Initialize time integrator for advection ---
        w_interp, iw_interp =\
            self.method[Interpolation].getWorkLengths(
                domain_dim=self.domain.dim)
        self._rw_interp = self._rwork[:w_interp]
        self._iw_interp = self._iwork[:iw_interp]

        vd = self.velocity.data[self.direction]
        num_interpolate = \
            self.method[Interpolation](vd, self.direction, topo,
                                       work=self._rw_interp,
                                       iwork=self._iw_interp)

        w_rk = self.method[TimeIntegrator].getWorkLengths(nb_components=1)
        self._rw_integ = self._rwork[w_interp:w_interp + w_rk]
        self.num_advec = self.method[TimeIntegrator](1, work=self._rw_integ,
                                                     f=num_interpolate,
                                                     topo=topo,
                                                     optim=WITH_GUESS)
        # --- Initialize remesh ---
        w_remesh, iw_remesh =\
            Remeshing.getWorkLengths(
                domain_dim=self.domain.dim)
        self._rw_remesh = self._rwork[:w_remesh]
        self._iw_remesh = self._iwork[:iw_remesh]

        self.num_remesh = Remeshing(self.method[Remesh],
                                    self.domain.dim,
                                    topo, self.direction,
                                    work=self._rw_remesh,
                                    iwork=self._iw_remesh)

        ## Particles positions
        start = max(w_interp + w_rk, w_remesh)
        self.part_position = [self._rwork[start]]

        ## Fields on particles
        self.fields_on_part = {}
        start += 1
        for f in self.advected_fields:
            self.fields_on_part[f] = self._rwork[start: start + f.nb_components]
            start += f.nb_components


    @debug
    @profile
    def apply(self, simulation, step_id):
        """
        Advection algorithm:
        - initialize particles and fields with their values on the grid.
        - compute particle positions in splitting direction,
        (time integrator), resolution of dx_p/dt = a_p.
        - remesh fields from particles to grid
        """
        assert simulation is not None, \
            'Simulation parameter is missing.'

        t, dt = simulation.time, simulation.time_step * dt_coeff
        # Initialize fields on particles with fields on grid values.
        for fg in self.advected_fields:
            for d in xrange(fg.nb_components):
                self.fields_on_part[fg][d][...] = fg[d][...]

        # Initialize particles on the grid
        toporef = self.advected_fields[0].topology
        self.part_position[0][...] = toporef.mesh.coords[self.direction]

        # Advect particles
        # RK use the first 2 (or 3) works and leave others to interpolation
        # First work contains fist evaluation of ode right hand side.
        self._rw_integ[0][...] = self.velocity.data[self.direction][...]
        self.part_position = self.num_advec(
            t, self.part_position, dt, result=self.part_position)

        # Remesh particles
        # It uses the last dim + 2 workspaces (same as interpolation)
        for fg in self.advected_fields:
            fp = self.fields_on_part[fg]
            for d in xrange(fg.nb_components):
                fg[d][...] = self.num_remesh(
                    self.part_position, fp[d], result=fg[d])


