# -*- coding: utf-8 -*-
"""Computation of the stretching term in Navier-Stokes.

See also
--------

* :ref:`stretching` in HySoP user guide.


"""
from hysop.constants import debug
from hysop.methods import TimeIntegrator, Formulation, \
    SpaceDiscretization
from hysop.numerics.finite_differences import FDC4
from hysop.operator.computational import Computational
from hysop.operator.continuous import opsetup
from hysop.operator.discrete.stretching import Conservative, GradUW
from hysop.operator.discrete.stretching import StretchingLinearized as SLD
from hysop.numerics.differential_operations import GradVxW, DivWV
from hysop.numerics.odesolvers import Euler


class Stretching(Computational):
    """Abstract interface to stretching operator.
    """

    _authorized_methods = [FDC4]
    _authorized_formulations = [Conservative, GradUW]

    @debug
    def __init__(self, velocity, vorticity, **kwds):
        """
        Parameters
        -----------
        velocity, vorticity : :class:`~hysop.fields.continuous_field.Field`
        **kwds : extra parameters for base class

        Notes
        -----
        * The default formulation is the 'Conservative' one.
        * The default solving method is finite differences, 4th order, in space
          and Runge-Kutta 3 in time.

        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(Stretching, self).__init__(variables=[velocity, vorticity],
                                         **kwds)
        # velocity variable (vector)
        self.velocity = velocity
        # vorticity variable (vector)
        self.vorticity = vorticity
        # Numerical methods for time and space discretization

        if self.method is None:
            import hysop.default_methods as default
            self.method = default.STRETCHING
        assert Formulation in self.method.keys()
        assert SpaceDiscretization in self.method.keys()
        assert TimeIntegrator in self.method.keys()
        msg = 'Stretching : unknown method for space discretization'
        assert self.method[SpaceDiscretization] in self._authorized_methods,\
            msg
        # Formulation used for the stretching equation.
        # Default = conservative form.
        self.formulation = self.method[Formulation]
        msg = 'Stretching error : unknown formulation.'
        assert self.formulation in self._authorized_formulations, msg
        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]

    def get_work_properties(self):
        super(Stretching, self).get_work_properties()
        # Get fields local shape.
        vd = self.discrete_fields[self.velocity]
        # Collect info from numerical methods
        # --> time-integrator required work space
        ti = self.method[TimeIntegrator]
        topo = vd.topology
        nbc = self.velocity.nb_components
        res = ti.get_work_properties(nbc, topo)
        # ---> differential operator work space
        if self.formulation is GradUW:
            dop = GradVxW

        elif self.formulation is Conservative:
            dop = DivWV
        res['rwork'] += dop.get_work_properties(topo)['rwork']
        return res

    @debug
    def discretize(self):
        nbghosts = self.method[SpaceDiscretization].ghosts_layer_size
        super(Stretching, self)._standard_discretize(nbghosts)

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        self.discrete_op =\
            self.formulation(velocity=self.discrete_fields[self.velocity],
                             vorticity=self.discrete_fields[self.vorticity],
                             method=self.method, rwork=rwork, iwork=iwork)
        self._is_uptodate = True


class StretchingLinearized(Stretching):
    """Solve the linearized stretching equation

    See details in :ref:`stretching`.

    """
    @debug
    def __init__(self, velocity_BF, vorticity_BF, **kwds):
        """
        Parameters
        -----------
        velocity_BF, vorticity_BF : base flow fields :
        class:`~hysop.fields.continuous_field.Field`
        **kwds : extra parameters for base class

        Notes
        -----
        * The default formulation is the 'Conservative' one.
        * The default solving method is finite differences, 4th order,
        in space and Runge-Kutta 3 in time.
        """
        super(StretchingLinearized, self).__init__(**kwds)
        self.variables[velocity_BF] = None
        self.variables[vorticity_BF] = None

        # Base flow velocity variable (vector)
        self.velocity_BF = velocity_BF
        # Base flow vorticity variable (vector)
        self.vorticity_BF = vorticity_BF
        # Usual stretching operator to compute the
        # first term of the linearized rhs: (w'.grad)ub
        self.usual_stretch = Stretching(self.velocity_BF, self.vorticity,
                                        discretization=self._discretization)

        self.input.append(self.velocity_BF)
        self.input.append(self.vorticity_BF)

    @debug
    def discretize(self):
        nbghosts = self.method[SpaceDiscretization].ghosts_layer_size
        self.usual_stretch.discretize()
        super(StretchingLinearized, self)._standard_discretize(nbghosts)

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        # Setup of the usual stretching operator
        self.usual_stretch.setup()

        # Setup of a second stretching operator (receiving 3 input variables)
        # to compute the 2nd term of the linearized rhs: (wb.grad)u'
        method_lin = self.method.copy()
        method_lin[TimeIntegrator] = Euler
        self.discrete_op =\
            SLD(velocity=self.discrete_fields[self.velocity],
                vorticity=self.discrete_fields[self.vorticity],
                vorticity_BF=self.discrete_fields[self.vorticity_BF],
                usual_op=self.usual_stretch.discrete_op,
                method=method_lin, rwork=rwork, iwork=iwork)
        self._is_uptodate = True
