# -*- coding: utf-8 -*-
from hysop import Field, Box
from hysop.operator.adapt_timestep import AdaptiveTimeStep
from hysop.problem.simulation import Simulation
from hysop.tools.parameters import Discretization
from hysop.core.mpi import main_comm, main_size
import numpy as np
from hysop.tools.numpywrappers import npw
from hysop.tools.misc import WorkSpaceTools
from hysop.numerics.tests.test_differential_operations import compute_vel,\
    compute_vort, diag_grad_func_3d, compute_max, grad_v_func_3d, strain_f_3d
from math import pi
from hysop.methods import TimeIntegrator

sin = np.sin
cos = np.cos
ldom = npw.asrealarray([pi * 2., 4. * pi, 2. * pi])
xdom = [0., 0., 0.]

d3d = Discretization([65, 123, 33], [2, 2, 2])


def init(criteria=None, work=False):
    """build fields and simu"""
    box = Box(length=ldom, origin=xdom)
    velo = Field(domain=box, formula=compute_vel,
                 name='Velocity', is_vector=True)
    vorti = Field(domain=box, formula=compute_vort,
                  name='Vorticity', is_vector=True)
    simu = Simulation(nb_iter=4)
    simu.initialize()
    assert simu.current_iteration == 0
    op = AdaptiveTimeStep(velo, vorti, simulation=simu,
                          criteria=criteria,
                          discretization=d3d, lcfl=0.125, cfl=0.5,
                          time_range=[0, 4])
    op.discretize()
    rwork = None
    if work is True:
        wkp = op.get_work_properties()
        lwork = len(wkp['rwork'])
        subshape = wkp['rwork'][0]
        rwork = WorkSpaceTools.check_work_array(lwork, subshape)
    op.setup(rwork=rwork)
    vorti.initialize(time=simu.time, topo=op.discrete_op.vorticity.topology)
    velo.initialize(time=simu.time, topo=op.discrete_op.vorticity.topology)
    op.apply()
    op.wait()
    return simu, op


def compute_dt_ref(topo, simu, op):
    """Compute some reference values for dt for
    each criteria
    """
    ref = Field(domain=topo.domain, formula=grad_v_func_3d,
                name='Analytical', nb_components=topo.domain.dim ** 2)
    rd = ref.discretize(topo)
    ref.initialize(time=simu.time, topo=topo)
    ref2 = Field(domain=topo.domain, formula=strain_f_3d,
                 name='Analytical',
                 nb_components=topo.domain.dim *
                 (topo.domain.dim + 1) / 2)
    rd2 = ref2.discretize(topo)
    ref2.initialize(time=simu.time, topo=topo)
    dd = topo.domain.dim
    ll = [rd[i * (dd + 1)] for i in xrange(dd)]
    ic = topo.mesh.compute_index
    refmax_adv = refmax_str = refmax_strain = 0.
    ll2 = [None, ] * dd
    ll2[0] = [rd2[0], rd2[3], rd2[4]]
    ll2[1] = [rd2[1], rd2[3], rd2[5]]
    ll2[2] = [rd2[2], rd2[4], rd2[5]]
    for d in xrange(dd):
        refmax_adv = max(compute_max([ll[d]], ic), refmax_adv)
        refmax_str = max(compute_max(rd[d * dd: (d + 1) * dd], ic), refmax_str)
        refmax_strain = max(compute_max(ll2[d], ic), refmax_strain)
    dt = {}
    dt['gradU'] = op.lcfl / refmax_adv
    dt['stretch'] = op.method[TimeIntegrator].stability_coeff() / refmax_str
    refmax = max([np.max(np.abs(v[ic]))
                  for v in op.velocity.discretize(topo).data])
    dt['cfl'] = op.cfl * topo.mesh.space_step[0] / refmax
    dt['strain'] = op.lcfl / refmax_strain
    refmax = max([np.max(np.abs(w[ic]))
                  for w in op.vorticity.discretize(topo).data])
    dt['vort'] = op.lcfl / refmax
    return dt


def test_adapt_vort():
    """
    Test 'vort' criteria (default config)
    """
    simu, op = init()
    assert 'vort' in op.criteria
    dt0 = simu.time_step
    topo = op.discrete_op.vorticity.topology
    wd = op.vorticity.discretize(topo)
    ic = topo.mesh.compute_index
    ref = max([np.max(np.abs(w[ic])) for w in wd.data])
    assert np.allclose(op.diagnostics()[2], op.lcfl / ref)
    assert np.allclose(simu.time_step, min(dt0, op.diagnostics()[2]))


def test_adapt_fail():
    """ Check wrong crit. value
    """
    fail = False
    try:
        init(['vorti', 'stretch'])
    except:
        fail = True
    assert fail


def test_adapt_gradu():
    """
    Test 'gradu' criteria
    """
    simu, op = init(['gradU'])
    dt0 = simu.time_step
    topo = op.discrete_op.velocity.topology
    ref = Field(domain=topo.domain, formula=diag_grad_func_3d,
                name='Analytical', is_vector=True)
    rd = ref.discretize(topo)
    ref.initialize(time=simu.time, topo=topo)
    dd = topo.domain.dim
    refmax = max([compute_max([rd[d]], topo.mesh.compute_index)
                  for d in xrange(dd)])
    assert np.allclose(op.diagnostics()[3], op.lcfl / refmax,
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(simu.time_step, min(dt0, op.diagnostics()[3]))


def test_adapt_strain():
    """
    Test 'strain' criteria
    """
    simu, op = init(['strain'])
    dt0 = simu.time_step
    topo = op.discrete_op.velocity.topology
    resdim = topo.domain.dim * (topo.domain.dim + 1) / 2
    ref = Field(domain=topo.domain, formula=strain_f_3d,
                name='Analytical', nb_components=resdim)
    rd = ref.discretize(topo)
    ref.initialize(time=simu.time, topo=topo)
    dd = topo.domain.dim
    ll = [None, ] * dd
    ll[0] = [rd[0], rd[3], rd[4]]
    ll[1] = [rd[1], rd[3], rd[5]]
    ll[2] = [rd[2], rd[4], rd[5]]
    ic = topo.mesh.compute_index
    refmax = 0.
    for d in xrange(dd):
        refmax = max(compute_max(ll[d], ic), refmax)
    dt = op.lcfl / refmax
    assert np.allclose(op.diagnostics()[4], dt,
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(simu.time_step, min(dt0, op.diagnostics()[4]))


def test_adapt_stretch():
    """
    Test 'stretch' criteria
    """
    simu, op = init(['stretch'])
    dt0 = simu.time_step
    topo = op.discrete_op.velocity.topology
    ref = Field(domain=topo.domain, formula=grad_v_func_3d,
                name='Analytical', nb_components=topo.domain.dim ** 2)
    rd = ref.discretize(topo)
    ref.initialize(time=simu.time, topo=topo)
    dd = topo.domain.dim
    refmax = 0.
    ic = topo.mesh.compute_index
    for d in xrange(dd):
        pos = d * dd
        refmax = max(compute_max(rd[pos: pos + dd], ic), refmax)
    cs = op.method[TimeIntegrator].stability_coeff()
    dt = cs / refmax
    assert np.allclose(op.diagnostics()[5], dt,
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(simu.time_step, min(dt0, op.diagnostics()[5]))


def test_adapt_cfl():
    """
    Test 'stretch' criteria
    """
    simu, op = init(['cfl'])
    dt0 = simu.time_step
    topo = op.discrete_op.velocity.topology
    dt0 = simu.time_step
    vd = op.velocity.discretize(topo)
    ic = topo.mesh.compute_index
    ref = max([np.max(np.abs(v[ic])) for v in vd.data])
    dt = op.cfl * topo.mesh.space_step[0] / ref
    assert np.allclose(op.diagnostics()[6], dt)
    assert np.allclose(simu.time_step, min(dt0, op.diagnostics()[6]))


def test_adapt_multi_crit():
    """
    Test combination of 3 criteria ('grad_u', 'stretch', 'cfl')
    """
    criteria = ['gradU', 'stretch', 'cfl']
    simu, op = init(criteria)
    topo = op.discrete_op.velocity.topology
    dtref = compute_dt_ref(topo, simu, op)
    dt0 = simu.time_step

    assert np.allclose(op.diagnostics()[3], dtref['gradU'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[5], dtref['stretch'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[6], dtref['cfl'],
                       rtol=topo.mesh.space_step.min())
    dtref_min = min([dtref[c] for c in criteria])
    assert np.allclose(op.diagnostics()[0], simu.time)
    assert np.allclose(op.diagnostics()[1], simu.time_step)
    assert simu.time_step <= op.maxdt
    assert np.allclose(simu.time_step, min(dt0, dtref_min))


def test_adapt_all_crit():
    """
    Test combination of all criteria
    """
    criteria = ['gradU', 'stretch', 'cfl', 'strain', 'vort']
    simu, op = init(criteria)
    topo = op.discrete_op.velocity.topology
    dtref = compute_dt_ref(topo, simu, op)
    dt0 = simu.time_step

    assert np.allclose(op.diagnostics()[0], simu.time)
    assert np.allclose(op.diagnostics()[1], simu.time_step)
    assert simu.time_step <= op.maxdt
    assert np.allclose(op.diagnostics()[2], dtref['vort'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[3], dtref['gradU'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[4], dtref['strain'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[5], dtref['stretch'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[6], dtref['cfl'],
                       rtol=topo.mesh.space_step.min())
    dtref_min = min([dtref[c] for c in criteria])
    assert np.allclose(op.diagnostics()[0], simu.time)
    assert np.allclose(op.diagnostics()[1], simu.time_step)
    assert simu.time_step <= op.maxdt
    assert np.allclose(simu.time_step, min(dt0, dtref_min))


def test_adapt_all_crit_with_work():
    """
    Test combination of all criteria
    """
    criteria = ['gradU', 'stretch', 'cfl', 'strain', 'vort']
    simu, op = init(criteria, work=True)
    topo = op.discrete_op.velocity.topology
    dtref = compute_dt_ref(topo, simu, op)
    dt0 = simu.time_step

    assert np.allclose(op.diagnostics()[0], simu.time)
    assert np.allclose(op.diagnostics()[1], simu.time_step)
    assert simu.time_step <= op.maxdt
    assert np.allclose(op.diagnostics()[2], dtref['vort'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[3], dtref['gradU'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[4], dtref['strain'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[5], dtref['stretch'],
                       rtol=topo.mesh.space_step.min())
    assert np.allclose(op.diagnostics()[6], dtref['cfl'],
                       rtol=topo.mesh.space_step.min())
    dtref_min = min([dtref[c] for c in criteria])
    assert np.allclose(op.diagnostics()[0], simu.time)
    assert np.allclose(op.diagnostics()[1], simu.time_step)
    assert simu.time_step <= op.maxdt
    assert np.allclose(simu.time_step, min(dt0, dtref_min))


def test_adapt_multi_tasks():
    """
    Multi mpi tasks version
    """
    # MPI procs are distributed among two tasks
    GPU = 4
    CPU = 1
    VISU = 12
    proc_tasks = [CPU, ] * main_size

    if main_size > 4:
        proc_tasks[-1] = GPU
        proc_tasks[2] = GPU
        proc_tasks[1] = VISU

    dom = Box(dimension=3, proc_tasks=proc_tasks)
    velo = Field(domain=dom, formula=compute_vel,
                 name='Velocity', is_vector=True)
    vorti = Field(domain=dom, formula=compute_vort,
                  name='Vorticity', is_vector=True)

    from hysop.tools.parameters import MPIParams
    cpu_task = MPIParams(comm=dom.task_comm, task_id=CPU)
    simu = Simulation(nb_iter=4)
    op = AdaptiveTimeStep(velo, vorti, simulation=simu, io_params=True,
                          discretization=d3d, lcfl=0.125, cfl=0.5,
                          mpi_params=cpu_task)
    simu.initialize()
    if dom.is_on_task(CPU):
        op.discretize()
        op.setup()
        vorti.initialize()

    while not simu.is_over:
        if dom.is_on_task(CPU):
            op.apply()
        op.wait()
        simu.advance()
        refval = 0
        if dom.is_on_task(CPU):
            refval = simu.time_step
        refval = main_comm.bcast(refval, root=0)
        assert refval == simu.time_step
