"""Test correction operator
"""
import hysop as pp
from hysop.operator.velocity_correction import VelocityCorrection
from hysop.problem.simulation import Simulation
import numpy as np
from hysop.tools.numpywrappers import npw
from hysop.tools.parameters import Discretization
from hysop.fields.tests.func_for_tests import v_TG, v_TG_2d, w_TG, w_TG_2d
pi = np.pi
cos = np.cos
sin = np.sin
# Upstream flow velocity
uinf = 1.0
tol = 1e-12


# Global resolution
g = 0
d2D = Discretization([33, 33], [g, g])
d3D = Discretization([33, 33, 33], [g, g, g])


def test_velocity_correction_3d():
    """Apply velocity correction, 3d domain
    """
    # Domain
    box = pp.Box(length=[2.0 * pi, pi, pi])
    # Vector Fields
    velo = pp.Field(domain=box, formula=v_TG,
                    name='Velocity', is_vector=True)
    vorti = pp.Field(domain=box, formula=w_TG,
                     name='Vorticity', is_vector=True)

    # Usual CartesianTopology topology definition
    topo = box.create_topology(discretization=d3D)

    ref_rate = npw.zeros(3)
    ref_rate[0] = uinf * box.length[1] * box.length[2]
    rate = pp.VariableParameter(data=ref_rate)
    op = VelocityCorrection(velo, vorti, req_flowrate=rate,
                            discretization=topo, io_params={})
    op.discretize()
    op.setup()
    # === Simulation setup ===
    simu = Simulation(start=0.0, end=5., time_step=0.005, max_iter=1000000)
    # init fields
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)
    # Apply correction
    op.apply(simu)
    # check new flowrate values
    sref = op.cb.surf[0]
    flowrate = sref.integrate_field_allc(velo, topo)
    assert (np.abs(flowrate - ref_rate) < tol).all()


def test_velocity_correction_2d():
    """Apply velocity correction, 2d domain
    """
    # Domain
    box = pp.Box(length=[2.0 * pi, pi], origin=[0., 0.])

    # Vector Fields
    velo = pp.Field(domain=box, formula=v_TG_2d,
                    name='Velocity', is_vector=True)
    vorti = pp.Field(domain=box, formula=w_TG_2d,
                     name='Vorticity', is_vector=False)

    # Usual CartesianTopology topology definition
    topo = box.create_topology(discretization=d2D)

    ref_rate = npw.zeros(2)
    ref_rate[0] = uinf * box.length[1]
    rate = pp.VariableParameter(data=ref_rate)
    op = VelocityCorrection(velo, vorti, req_flowrate=rate,
                            discretization=topo)
    op.discretize()
    op.setup()
    # === Simulation setup ===
    simu = Simulation(start=0.0, end=5., time_step=0.005, max_iter=1000000)
    # init fields
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)

    # Apply correction
    op.apply(simu)
    # check new flowrate values
    sref = op.cb.surf[0]
    flowrate = sref.integrate_field_allc(velo, topo)
    assert (np.abs(flowrate - ref_rate) < tol).all()
