import os
import shutil
import datetime
import inspect
import argparse
import sys
import numpy as np

from hysop import MPI
from hysop.tools.types import check_instance
from hysop.fields.discrete_field import DiscreteScalarFieldView


class DebugDumper(object):
    def __init__(self, name, path='/tmp/hysop/debug', force_overwrite=False,
                 enable_on_op_apply=False, dump_precision=10,
                 comm=MPI.COMM_WORLD, io_leader=0):
        directory = path+'/'+name
        blobs_directory = directory + '/data'

        self.name = name
        self.directory = directory
        self.blobs_directory = blobs_directory
        self.dump_id = 0
        self.enable_on_op_apply = enable_on_op_apply
        self.dump_precision = dump_precision

        self.comm = comm
        self.io_leader = io_leader
        self.comm_size = comm.Get_size()
        self.comm_rank = comm.Get_rank()
        assert 0 <= self.io_leader < self.comm_size
        self.is_io_leader = (self.comm_rank == self.io_leader)

        if self.is_io_leader:
            if os.path.exists(directory):
                if force_overwrite:
                    shutil.rmtree(directory)
                else:
                    msg = 'Directory \'{}\' already exists.'.format(directory)
                    raise RuntimeError(msg)
            os.makedirs(blobs_directory)
            runfile = '{}/run.txt'.format(directory)
            runfile = open(runfile, 'a')
            self.runfile = runfile
            self.print_header()
        else:
            self.runfile = None

    def __del__(self):
        if hasattr(self, 'runfile') and (self.runfile is not None):
            self.runfile.close()
            self.runfile = None

    @classmethod
    def lformat(cls, id_, iteration, time, tag, min_, max_, mean, variance, dtype, shape, description='', dump_precision=None):
        try:
            return '{:<4}  {:<10}  {:<20.{p}f}  {:<40}  {:<+20.{p}f}  {:<+20.{p}f}  {:<+20.{p}f}  {:<+20.{p}f}  {:<20}  {:<20}  {}'.format(
                id_, iteration, time, tag, min_, max_, mean, variance, dtype, shape, description, p=dump_precision)
        except:
            return '{:<4}  {:<10}  {:<20}  {:<40}  {:<20}  {:<20}  {:<20}  {:<20}  {:<20}  {:<20}  {}'.format(
                id_, iteration, time, tag, min_, max_, mean, variance, dtype, shape, description)

    def print_header(self):
        now = datetime.datetime.now()
        self.runfile.write('DEBUG DUMP {} ({})\n'.format(
            self.name,
            now.strftime("%Y-%m-%d %H:%M")))
        self.runfile.write(self.lformat('id', 'iteration', 'time', 'tag', 'min', 'max', 'mean', 'variance',
                                        'dtype', 'shape', 'description'))

    @classmethod
    def get_arrays(self, arrays):
        if isinstance(arrays, DiscreteScalarFieldView):
            return tuple(d.get()[arrays.compute_slices] for d in arrays.data)
        elif isinstance(arrays, np.ndarray):
            return (arrays,)
        else:
            check_instance(arrays, tuple, values=np.ndarray)
            return arrays

    def __call__(self, iteration, time, tag, arrays, description=''):
        check_instance(iteration, int)
        arrays = self.get_arrays(arrays)
        N = len(arrays)
        comm = self.comm
        comm_size = self.comm_size
        for (i, data) in enumerate(arrays):
            if (N > 1):
                tag_ = '{}_{}'.format(tag, i)
            else:
                tag_ = tag
            if (description != ''):
                if (N > 1):
                    description_ = '{} (component {})'.format(description, i)
                else:
                    description_ = description
            else:
                _file, _line = inspect.stack()[1][1:3]
                description_ = '{}:{}'.format(_file, _line)
            dtype = data.dtype
            if dtype in (np.complex64, np.complex128):
                data = (data.real, data.imag)
                tags = (tag_+'_real', tag_+'_imag')
            else:
                data = (data,)
                tags = (tag_,)
            for (d, tag_) in zip(data, tags):
                dtype = d.dtype
                shape = None
                id_ = self.dump_id
                _min = comm.allreduce(float(np.nanmin(d)),  op=MPI.MIN)
                _max = comm.allreduce(float(np.nanmax(d)),  op=MPI.MAX)
                mean = comm.allreduce(float(np.nanmean(d))) / comm_size
                variance = comm.allreduce(float(np.nansum((d-mean)**2))) / \
                    float(comm.allreduce(long(d.size)))
                entry = '\n'+self.lformat(id_, iteration, time, tag_, _min, _max,
                                          mean, variance, dtype, shape, description_, self.dump_precision)

                if self.is_io_leader:
                    self.runfile.write(entry)

                if (comm_size == 1):
                    dst = '{}/{}'.format(self.blobs_directory, self.dump_id)
                    np.savez_compressed(dst, data=d)

                self.dump_id += 1
